from datetime import datetime, timedelta
from collections import Counter
from copy import deepcopy
from functools import reduce
from funcy import curry
import pandas as pd

from cadCAD.configuration.utils.depreciationHandler import sanitize_partial_state_updates
from cadCAD.utils import dict_filter, contains_type, flatten_tabulated_dict, tabulate_dict


class TensorFieldReport:
    def __init__(self, config_proc):
        self.config_proc = config_proc

    def create_tensor_field(self, partial_state_updates, exo_proc, keys=['policies', 'variables']):

        partial_state_updates = sanitize_partial_state_updates(partial_state_updates) # Temporary

        dfs = [self.config_proc.create_matrix_field(partial_state_updates, k) for k in keys]
        df = pd.concat(dfs, axis=1)
        for es, i in zip(exo_proc, range(len(exo_proc))):
            df['es' + str(i + 1)] = es
        df['m'] = df.index + 1
        return df


def configs_as_spec(configs):
    sim_ids = list(map(lambda x: x.simulation_id, configs))
    sim_id_counts = list(Counter(sim_ids).values())
    IDed_configs = list(zip(sim_ids, configs))
    del sim_ids
    selected_IDed_configs = dict(IDed_configs)
    del IDed_configs
    counted_IDs_configs = list(zip(sim_id_counts, selected_IDed_configs.values()))
    del sim_id_counts, selected_IDed_configs
    for runs, config in counted_IDs_configs:
        config.sim_config['N'] = runs
    return counted_IDs_configs


def configs_as_objs(configs):
    counted_IDs_configs = configs_as_spec(configs)
    new_configs = list(map(lambda x: x[1], counted_IDs_configs))
    del counted_IDs_configs
    return new_configs


def configs_as_dicts(configs):
    counted_IDs_configs = configs_as_spec(configs)
    new_configs = list(map(lambda x: x[1].__dict__, counted_IDs_configs))
    del counted_IDs_configs
    return new_configs


def configs_as_dataframe(configs):
    new_configs = configs_as_dicts(configs)
    configs_df = pd.DataFrame(new_configs)
    del new_configs
    configs_df_columns = list(configs_df.columns)
    header_cols = ['session_id', 'user_id', 'simulation_id', 'run_id']
    for col in header_cols:
        configs_df_columns.remove(col)
    return configs_df[header_cols + configs_df_columns]

## System Model

def state_update(y, x):
    return lambda var_dict, sub_step, sL, s, _input, **kwargs: (y, x)


def bound_norm_random(rng, low, high):
    res = rng.normal((high+low)/2, (high-low)/6)
    if res < low or res > high:
        res = bound_norm_random(rng, low, high)
    # return Decimal(res)
    return float(res)


tstep_delta = timedelta(days=0, minutes=0, seconds=30)
def time_step(dt_str, dt_format='%Y-%m-%d %H:%M:%S', _timedelta = tstep_delta):
    dt = datetime.strptime(dt_str, dt_format)
    t = dt + _timedelta
    return t.strftime(dt_format)


ep_t_delta = timedelta(days=0, minutes=0, seconds=1)
def ep_time_step(s_condition, dt_str, fromat_str='%Y-%m-%d %H:%M:%S', _timedelta = ep_t_delta):
    if s_condition:
        return time_step(dt_str, fromat_str, _timedelta)
    else:
        return dt_str


def exo_update_per_ts(ep):
    # @curried
    def ep_decorator(f, y, var_dict, sub_step, sL, s, _input,  **kwargs):
        if s['substep'] + 1 == 1:
            return f(var_dict, sub_step, sL, s, _input,  **kwargs)
        else:
            return y, s[y]

    return {es: ep_decorator(f, es) for es, f in ep.items()}


def trigger_condition(s, pre_conditions, cond_opp):
    condition_bools = [s[field] in precondition_values for field, precondition_values in pre_conditions.items()]
    return reduce(cond_opp, condition_bools)


def apply_state_condition(pre_conditions, cond_opp, y, f, _g, step, sL, s, _input, **kwargs):
    def state_scope_tuner(f):
        lenf = f.__code__.co_argcount
        if lenf == 5:
            return f(_g, step, sL, s, _input)
        elif lenf == 6:
            return f(_g, step, sL, s, _input, **kwargs)

    if trigger_condition(s, pre_conditions, cond_opp):
        return state_scope_tuner(f)
    else:
        return y, s[y]


def var_trigger(y, f, pre_conditions, cond_op):
    return lambda _g, step, sL, s, _input, **kwargs: \
        apply_state_condition(pre_conditions, cond_op, y, f, _g, step, sL, s, _input, **kwargs)


def var_substep_trigger(substeps):
    def trigger(end_substep, y, f):
        pre_conditions = {'substep': substeps}
        cond_opp = lambda a, b: a and b
        return var_trigger(y, f, pre_conditions, cond_opp)

    return lambda y, f: curry(trigger)(substeps)(y)(f)


def env_trigger(end_substep):
    def trigger(end_substep, trigger_field, trigger_vals, funct_list):
        def env_update(state_dict, sweep_dict, target_value):
            state_dict_copy = deepcopy(state_dict)
            # Use supstep to simulate current sysMetrics
            if state_dict_copy['substep'] == end_substep:
                state_dict_copy['timestep'] = state_dict_copy['timestep'] + 1

            if state_dict_copy[trigger_field] in trigger_vals:
                for g in funct_list:
                    target_value = g(sweep_dict, target_value)

            del state_dict_copy
            return target_value

        return env_update

    return lambda trigger_field, trigger_vals, funct_list: \
        curry(trigger)(end_substep)(trigger_field)(trigger_vals)(funct_list)


def config_sim(d):
    def process_variables(d):
        return flatten_tabulated_dict(tabulate_dict(d))

    if "M" in d:
        return [{"N": d["N"], "T": d["T"], "M": M} for M in process_variables(d["M"])]
    else:
        d["M"] = [{}]
        return d


def psub_list(psu_block, psu_steps):
    return [psu_block[psu] for psu in psu_steps]


def psub(policies, state_updates):
    return {
        'policies': policies,
        'states': state_updates
    }


def genereate_psubs(policy_grid, states_grid, policies, state_updates):
    PSUBS = []
    for policy_ids, state_list in zip(policy_grid, states_grid):
        filtered_policies = {k: v for (k, v) in policies.items() if k in policy_ids}
        filtered_state_updates = {k: v for (k, v) in state_updates.items() if k in state_list}
        PSUBS.append(psub(filtered_policies, filtered_state_updates))

    return PSUBS


def access_block(state_history, target_field, psu_block_offset, exculsion_list=[]):
    exculsion_list += [target_field]

    def filter_history(key_list, block):
        filter = lambda key_list: \
            lambda d: {k: v for k, v in d.items() if k not in key_list}
        return list(map(filter(key_list), block))

    if psu_block_offset < -1:
        if len(state_history) >= abs(psu_block_offset):
            return filter_history(exculsion_list, state_history[psu_block_offset])
        else:
            return []
    elif psu_block_offset == -1:
        return filter_history(exculsion_list, state_history[psu_block_offset])
    else:
        return []

## Parameter Sweep
def partial_state_sweep_filter(state_field, partial_state_updates):
    partial_state_dict = dict([(k, v[state_field]) for k, v in partial_state_updates.items()])
    return dict([
        (k, dict_filter(v, lambda v: isinstance(v, list))) for k, v in partial_state_dict.items()
            if contains_type(list(v.values()), list)
    ])


def state_sweep_filter(raw_exogenous_states):
    return dict([(k, v) for k, v in raw_exogenous_states.items() if isinstance(v, list)])


# @curried
def sweep_partial_states(_type, in_config):
    configs = []
    # filtered_mech_states
    filtered_partial_states = partial_state_sweep_filter(_type, in_config.partial_state_updates)
    if len(filtered_partial_states) > 0:
        for partial_state, state_dict in filtered_partial_states.items():
            for state, state_funcs in state_dict.items():
                for f in state_funcs:
                    config = deepcopy(in_config)
                    config.partial_state_updates[partial_state][_type][state] = f
                    configs.append(config)
                    del config
    else:
        configs = [in_config]

    return configs

# @curried
def sweep_states(state_type, states, in_config):
    configs = []
    filtered_states = state_sweep_filter(states)
    if len(filtered_states) > 0:
        for state, state_funcs in filtered_states.items():
            for f in state_funcs:
                config = deepcopy(in_config)
                exploded_states = deepcopy(states)
                exploded_states[state] = f
                if state_type == 'exogenous':
                    config.exogenous_states = exploded_states
                elif state_type == 'environmental':
                    config.env_processes = exploded_states
                configs.append(config)
                del config, exploded_states
    else:
        configs = [in_config]

    return configs

# @curried
# def env_proc_trigger(timestep, f, time):
#     if time == timestep:
#         return f
#     else:
#         return lambda x: x



def get_base_value(x):
    if isinstance(x, str):
        return ''
    elif isinstance(x, int):
        return 0
    elif isinstance(x, list):
        return []
    else:
        return 0


def policy_to_dict(v):
    return dict(list(zip(map(lambda n: 'p' + str(n + 1), list(range(len(v)))), v)))


add = lambda a, b: a + b


# @curried
# def foldr_dict_vals(f, d):
#     return foldr(f)(list(d.values()))
#
#
# def sum_dict_values():
#     return foldr_dict_vals(add)


# @curried
# def dict_op(f, d1, d2):
#     def set_base_value(target_dict, source_dict, key):
#         if key not in target_dict:
#             return get_base_value(source_dict[key])
#         else:
#             return target_dict[key]
#
#     key_set = set(list(d1.keys()) + list(d2.keys()))
#
#     return {k: f(set_base_value(d1, d2, k), set_base_value(d2, d1, k)) for k in key_set}
#
# def dict_elemwise_sum():
#     return dict_op(add)



from typing import Dict, Callable, List, Tuple
from pandas.core.frame import DataFrame
from collections import deque
from copy import deepcopy
import pandas as pd

from cadCAD import configs
from cadCAD.utils import key_filter
from cadCAD.configuration.utils import exo_update_per_ts
from cadCAD.configuration.utils.depreciationHandler import sanitize_partial_state_updates, sanitize_config


class Configuration(object):
    def __init__(self, user_id, subset_id, subset_window, sim_config={}, initial_state={}, seeds={}, env_processes={},
                 exogenous_states={}, partial_state_update_blocks={}, policy_ops=[lambda a, b: a + b],
                 session_id=0, simulation_id=0, run_id=1, experiment_id=0, exp_window=deque([0, None], 2), **kwargs
    ) -> None:
        self.sim_config = sim_config
        self.initial_state = initial_state
        self.seeds = seeds
        self.env_processes = env_processes
        self.exogenous_states = exogenous_states
        self.partial_state_updates = partial_state_update_blocks
        self.policy_ops = policy_ops
        self.kwargs = kwargs

        self.user_id = user_id
        self.session_id = session_id # essentially config id
        self.simulation_id = simulation_id
        self.run_id = run_id
        self.experiment_id = experiment_id
        self.exp_window = exp_window
        self.subset_id = subset_id
        self.subset_window = subset_window

        sanitize_config(self)


class Experiment:
    def __init__(self):
        self.exp_id = 0
        self.subset_id = 0
        self.exp_window = deque([self.exp_id, None], 2)
        self.subset_window = deque([self.subset_id, None], 2)

    def append_configs(
            self,
            user_id='cadCAD_user',
            sim_configs={}, initial_state={}, seeds={}, raw_exogenous_states={}, env_processes={},
            partial_state_update_blocks={}, policy_ops=[lambda a, b: a + b], _exo_update_per_ts: bool = True,
            config_list=configs
    ) -> None:

        try:
            max_runs = sim_configs[0]['N']
        except KeyError:
            max_runs = sim_configs['N']

        if _exo_update_per_ts is True:
            exogenous_states = exo_update_per_ts(raw_exogenous_states)
        else:
            exogenous_states = raw_exogenous_states

        if isinstance(sim_configs, dict):
            sim_configs = [sim_configs]

        simulation_id = 0
        if len(config_list) > 0:
            last_config = config_list[-1]
            simulation_id = last_config.simulation_id + 1

        sim_cnt = 0
        new_sim_configs = []
        for subset_id, t in enumerate(list(zip(sim_configs, list(range(len(sim_configs)))))):
            sim_config = t[0]
            sim_config['subset_id'] = subset_id
            sim_config['subset_window'] = self.subset_window
            N = sim_config['N']
            if N > 1:
                for n in range(N):
                    sim_config['simulation_id'] = simulation_id + sim_cnt
                    sim_config['run_id'] = n
                    sim_config['N'] = 1
                    # sim_config['N'] = n + 1
                    new_sim_configs.append(deepcopy(sim_config))
                del sim_config
            else:
                sim_config['simulation_id'] = simulation_id
                sim_config['run_id'] = 0
                new_sim_configs.append(deepcopy(sim_config))
                # del sim_config

            sim_cnt += 1

        run_id = 0
        for sim_config in new_sim_configs:
            subset_id = sim_config['subset_id']
            sim_config['N'] = run_id + 1
            if max_runs == 1:
                sim_config['run_id'] = run_id
            elif max_runs >= 1:
                if run_id >= max_runs:
                    sim_config['N'] = run_id - (max_runs - 1)

            self.exp_window = deepcopy(self.exp_window)
            config = Configuration(
                sim_config=sim_config,
                initial_state=initial_state,
                seeds=seeds,
                exogenous_states=exogenous_states,
                env_processes=env_processes,
                partial_state_update_blocks=partial_state_update_blocks,
                policy_ops=policy_ops,

                # session_id=session_id,
                user_id=user_id,
                session_id=f"{user_id}={sim_config['simulation_id']}_{sim_config['run_id']}",
                simulation_id=sim_config['simulation_id'],
                run_id=sim_config['run_id'],

                experiment_id=self.exp_id,
                exp_window=self.exp_window,
                subset_id=subset_id,
                subset_window=self.subset_window
            )
            configs.append(config)
            run_id += 1
        self.exp_id += 1
        self.exp_window.appendleft(self.exp_id)


class Identity:
    def __init__(self, policy_id: Dict[str, int] = {'identity': 0}) -> None:
        self.beh_id_return_val = policy_id

    def p_identity(self, var_dict, sub_step, sL, s, **kwargs):
        return self.beh_id_return_val

    def policy_identity(self, k: str) -> Callable:
        return self.p_identity

    def no_state_identity(self, var_dict, sub_step, sL, s, _input, **kwargs):
        return None

    def state_identity(self, k: str) -> Callable:
        return lambda var_dict, sub_step, sL, s, _input, **kwargs: (k, s[k])

    # state_identity = cloudpickle.dumps(state_identity)

    def apply_identity_funcs(self,
                             identity: Callable,
                             df: DataFrame,
                             cols: List[str]) -> DataFrame:
        """
        Apply the identity on each df column, using its self value as the
        argument.
        """
        fill_values = {col: identity(col) for col in cols}
        filled_df = df.fillna(fill_values)
        return filled_df


class Processor:
    def __init__(self, id: Identity = Identity()) -> None:
        self.id = id
        self.p_identity = id.p_identity
        self.policy_identity = id.policy_identity
        self.no_state_identity = id.no_state_identity
        self.state_identity = id.state_identity
        self.apply_identity_funcs = id.apply_identity_funcs

    def create_matrix_field(self, partial_state_updates, key: str) -> DataFrame:
        if key == 'variables':
            identity = self.state_identity
        elif key == 'policies':
            identity = self.policy_identity

        df = pd.DataFrame(key_filter(partial_state_updates, key))
        filled_df = self.apply_identity_funcs(identity, df, list(df.columns))
        if len(filled_df) > 0:
            return filled_df
        else:
            return pd.DataFrame({'empty': []})

    def generate_config(self, initial_state, partial_state_updates, exo_proc
                       ) -> List[Tuple[List[Callable], List[Callable]]]:

        def no_update_handler(bdf, sdf):
            if (bdf.empty == False) and (sdf.empty == True):
                bdf_values = bdf.values.tolist()
                sdf_values = [[self.no_state_identity] * len(bdf_values) for m in range(len(partial_state_updates))]
                return sdf_values, bdf_values
            elif (bdf.empty == True) and (sdf.empty == False):
                sdf_values = sdf.values.tolist()
                bdf_values = [[self.p_identity] * len(sdf_values) for m in range(len(partial_state_updates))]
                return sdf_values, bdf_values
            else:
                sdf_values = sdf.values.tolist()
                bdf_values = bdf.values.tolist()
                return sdf_values, bdf_values

        def only_ep_handler(state_dict):
            sdf_functions = [
                lambda var_dict, sub_step, sL, s, _input, **kwargs: (k, v) for k, v in zip(state_dict.keys(), state_dict.values())
            ]
            sdf_values = [sdf_functions]
            bdf_values = [[self.p_identity] * len(sdf_values)]
            return sdf_values, bdf_values

        if len(partial_state_updates) != 0:
            # backwards compatibility
            partial_state_updates = sanitize_partial_state_updates(partial_state_updates)

            bdf = self.create_matrix_field(partial_state_updates, 'policies')
            sdf = self.create_matrix_field(partial_state_updates, 'variables')
            sdf_values, bdf_values = no_update_handler(bdf, sdf)
            zipped_list = list(zip(sdf_values, bdf_values))
        else:
            sdf_values, bdf_values = only_ep_handler(initial_state)
            zipped_list = list(zip(sdf_values, bdf_values))

        return list(map(lambda x: (x[0] + exo_proc, x[1]), zipped_list))



from pprint import pprint
import sys
import pandas as pd
from tabulate import tabulate

from cadCAD.utils import arrange_cols

from simulations.regression_tests.poc.action_policies import enter_action, message_actions, exit_action
from simulations.regression_tests.poc.state_updates import send_message, count_messages, add_send_time, current_time

# State Updates
from simulations.regression_tests.experiments import poc

variables = {
    'client_a': send_message('client_a'),
    'client_b': send_message('client_b'),
    'total_msg_count': count_messages,
    'total_send_time': add_send_time,
    'record_creation': current_time('record_creation')
}

# Action Policies
policy_group_1 = {
    "action_1": enter_action('server', 'room_1', 'A'),
    "action_2": enter_action('server', 'room_1', 'B')
}

policy_group_2 = {
    "action_1": message_actions('client_A', 'room_1', "Hi B", 'A', 'B'),
    "action_2": message_actions('client_B', 'room_1', "Hi A", 'B', 'A')
}

policy_group_3 = {
    "action_1": message_actions('client_A', 'room_1', "Bye B", 'A', 'B'),
    "action_2": message_actions('client_B', 'room_1', "Bye A", 'B', 'A')
}

policy_group_4 = {
    "action_1": exit_action('server', 'room_1', 'A'),
    "action_2": exit_action('server', 'room_1', 'B')
}

policy_groups = [policy_group_1, policy_group_2, policy_group_3, policy_group_4]
# Partial State update Block
sim_composition = [{'policies': policy_group, 'variables': variables} for policy_group in policy_groups]


def main(executor, sim_config, intitial_conditions, sim_composition):
    poc.append_configs(
        user_id='Joshua',
        sim_configs=sim_config,
        initial_state=intitial_conditions,
        partial_state_update_blocks=sim_composition,
        policy_ops=[lambda a, b: a + b],
        _exo_update_per_ts=False
    )
    pprint(sim_composition)

    rdd, tensor_fields, sessions = executor.execute()
    rdd_result = rdd.collect()
    result = pd.DataFrame(rdd_result)
    print(tabulate(tensor_fields[0], headers='keys', tablefmt='psql'))
    pprint(sessions)
    print(tabulate(result.head(100), headers='keys', tablefmt='psql'))

    # i = 0
    # lasts = []
    # mem_size = 0
    # result_count = 0
    # for raw_result, tensor_field, session in executor.execute():
    #     result = pd.DataFrame(raw_result)
    #     # sesh = result[['user_id', 'session_id', 'simulation_id', 'run_id', 'run', 'timestep', 'substep']]
    #     # print(tabulate(sesh, headers='keys', tablefmt='psql'))
    #     # exit()
    #     metrics_result = result[
    #         [
    #             'timestep', 'substep',
    #             'record_creation', 'total_msg_count', 'total_send_time'
    #         ]
    #     ]
    #     msgs_result = result[
    #         [
    #             'timestep', 'substep',
    #             'record_creation',
    #             #'client_a', 'client_b'
    #         ]
    #     ]
    #     print()
    #     result_len = len(result.index)
    #     if i == 0:
    #         print()
    #         print(tabulate(tensor_field, headers='keys', tablefmt='psql'))
    #         # print(session)
    #         # print(f'result_len: {result_len}')
    #         # print(tabulate(msgs_result.tail(5), headers='keys', tablefmt='psql'))
    #         print()
    #
    #     # print(session)
    #     # print(tabulate(sesh.head(5), headers='keys', tablefmt='psql'))
    #     print(session)
    #     # result_len = len(result.index)
    #     # print(result_len)
    #     # mem_size += sys.getsizeof(result)
    #     result_count += result_len
    #     print(tabulate(msgs_result.head(5), headers='keys', tablefmt='psql'))
    #     # print()
    #     i += 1
    # # print(mem_size / 1073741824)
    # print("df total result: " + str(result_count))



import pprint
from typing import Dict, List

# from cadCAD.configuration import append_configs
from cadCAD.configuration.utils import env_trigger, var_substep_trigger, config_sim, psub_list
from testing.experiments import exp_param_sweep

pp = pprint.PrettyPrinter(indent=4)

def some_function(x):
    return x

# Optional
# dict must contain lists opf 2 distinct lengths
g: Dict[str, List[int]] = {
    'alpha': [1],
    'beta': [2, some_function],
    'gamma': [3, 4],
    # 'beta': [1],
    # 'gamma': [4],
    'omega': [7]
}

psu_steps = ['m1', 'm2', 'm3']
system_substeps = len(psu_steps)
var_timestep_trigger = var_substep_trigger([0, system_substeps])
env_timestep_trigger = env_trigger(system_substeps)
env_process = {}


# ['s1', 's2', 's3', 's4']
# Policies per Mechanism
def gamma(_g, step, sL, s, **kwargs):
    return {'gamma': _g['gamma']}


def omega(_g, step, sL, s, **kwargs):
    return {'omega': _g['omega']}


# Internal States per Mechanism
def alpha(_g, step, sL, s, _input, **kwargs):
    return 'alpha', _g['alpha']


def beta(_g, step, sL, s, _input, **kwargs):
    return 'beta', _g['beta']


def policies(_g, step, sL, s, _input, **kwargs):
    return 'policies', _input


def sweeped(_g, step, sL, s, _input, **kwargs):
    return 'sweeped', {'beta': _g['beta'], 'gamma': _g['gamma']}

psu_block = {k: {"policies": {}, "variables": {}} for k in psu_steps}
for m in psu_steps:
    psu_block[m]['policies']['gamma'] = gamma
    psu_block[m]['policies']['omega'] = omega
    psu_block[m]["variables"]['alpha'] = alpha
    psu_block[m]["variables"]['beta'] = beta
    psu_block[m]['variables']['policies'] = policies
    psu_block[m]["variables"]['sweeped'] = var_timestep_trigger(y='sweeped', f=sweeped)


# Genesis States
genesis_states = {
    'alpha': 0,
    'beta': 0,
    'policies': {},
    'sweeped': {}
}

# Environment Process
env_process['sweeped'] = env_timestep_trigger(trigger_field='timestep', trigger_vals=[5], funct_list=[lambda _g, x: _g['beta']])


sim_config = config_sim(
    {
        "N": 2,
        "T": range(5),
        "M": g, # Optional
    }
)

# New Convention
partial_state_update_blocks = psub_list(psu_block, psu_steps)
exp_param_sweep.append_configs(
    sim_configs=sim_config,
    initial_state=genesis_states,
    env_processes=env_process,
    partial_state_update_blocks=partial_state_update_blocks
)



import numpy as np
from datetime import timedelta

from cadCAD.configuration.utils import bound_norm_random, config_sim, time_step, env_trigger
from cadCAD.configuration import Experiment

seeds = {
    'z': np.random.RandomState(1),
    'a': np.random.RandomState(2),
    'b': np.random.RandomState(3),
    'c': np.random.RandomState(4)
}


# Policies per Mechanism
def p1m1(_g, step, sH, s):
    return {'param1': 1}
def p2m1(_g, step, sH, s):
    return {'param1': 1, 'param2': 4}

def p1m2(_g, step, sH, s):
    return {'param1': 'a', 'param2': 2}
def p2m2(_g, step, sH, s):
    return {'param1': 'b', 'param2': 4}

def p1m3(_g, step, sH, s):
    return {'param1': ['c'], 'param2': np.array([10, 100])}
def p2m3(_g, step, sH, s):
    return {'param1': ['d'], 'param2': np.array([20, 200])}


# Internal States per Mechanism
def s1m1(_g, step, sH, s, _input):
    y = 's1'
    x = s['s1'] + 1
    return (y, x)
def s2m1(_g, step, sH, s, _input):
    y = 's2'
    x = _input['param2']
    return (y, x)

def s1m2(_g, step, sH, s, _input):
    y = 's1'
    x = s['s1'] + 1
    return (y, x)
def s2m2(_g, step, sH, s, _input):
    y = 's2'
    x = _input['param2']
    return (y, x)

def s1m3(_g, step, sH, s, _input):
    y = 's1'
    x = s['s1'] + 1
    return (y, x)
def s2m3(_g, step, sH, s, _input):
    y = 's2'
    x = _input['param2']
    return (y, x)

def policies(_g, step, sH, s, _input):
    y = 'policies'
    x = _input
    return (y, x)


# Exogenous States
proc_one_coef_A = 0.7
proc_one_coef_B = 1.3

def es3(_g, step, sH, s, _input):
    y = 's3'
    x = s['s3'] * bound_norm_random(seeds['a'], proc_one_coef_A, proc_one_coef_B)
    return (y, x)

def es4(_g, step, sH, s, _input):
    y = 's4'
    x = s['s4'] * bound_norm_random(seeds['b'], proc_one_coef_A, proc_one_coef_B)
    return (y, x)

def update_timestamp(_g, step, sH, s, _input):
    y = 'timestamp'
    return y, time_step(dt_str=s[y], dt_format='%Y-%m-%d %H:%M:%S', _timedelta=timedelta(days=0, minutes=0, seconds=1))


# Genesis States
genesis_states = {
    's1': 0.0,
    's2': 0.0,
    's3': 1.0,
    's4': 1.0,
    'timestamp': '2018-10-01 15:16:24'
}


# Environment Process
trigger_timestamps = ['2018-10-01 15:16:25', '2018-10-01 15:16:27', '2018-10-01 15:16:29']
env_processes = {
    "s3": [lambda _g, x: 5],
    "s4": env_trigger(3)(trigger_field='timestamp', trigger_vals=trigger_timestamps, funct_list=[lambda _g, x: 10])
}


psubs = [
    {
        "policies": {
            "b1": p1m1,
            "b2": p2m1
        },
        "variables": {
            "s1": s1m1,
            "s2": s2m1,
            "s3": es3,
            "s4": es4,
            "timestamp": update_timestamp
        }
    },
    {
        "policies": {
            "b1": p1m2,
            "b2": p2m2
        },
        "variables": {
            "s1": s1m2,
            "s2": s2m2,
            # "s3": es3p1,
            # "s4": es4p2,
        }
    },
    {
        "policies": {
            "b1": p1m3,
            "b2": p2m3
        },
        "variables": {
            "s1": s1m3,
            "s2": s2m3,
            # "s3": es3p1,
            # "s4": es4p2,
        }
    }
]


sim_config = config_sim(
    {
        "N": 2,
        "T": range(1),
    }
)

exp = Experiment()
exp.append_configs(
    sim_configs=sim_config,
    initial_state=genesis_states,
    env_processes=env_processes,
    partial_state_update_blocks=psubs,
    policy_ops=[lambda a, b: a + b]
)


from cadCAD import logo


def print_exec_info(exec_context, configs):
    print(logo)
    print(f'Execution Mode: {exec_context}')
    print(f'Configuration Count: {len(configs)}')
    first_sim = configs[0].sim_config
    n_t = len(first_sim['T'])
    n_m = len(first_sim['M'])
    n_n = first_sim['N']
    n_s = len(configs[0].initial_state)
    print(f'Dimensions of the first simulation: (Timesteps, Params, Runs, Vars) = ({n_t}, {n_m}, {n_n}, {n_s})')


import unittest
import pandas as pd

from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from testing.generic_test import make_generic_test
from testing.models import policy_aggregation
from cadCAD import configs

exec_mode = ExecutionMode()
exec_ctx = ExecutionContext(context=exec_mode.local_mode)
run = Executor(exec_context=exec_ctx, configs=configs)

raw_result, tensor_field, _ = run.execute()
result = pd.DataFrame(raw_result)

expected_results = {
    (0, 1, 0, 0): {'policies': {}, 's1': 0},
    (0, 1, 1, 1): {'policies': {'policy1': 2, 'policy2': 4}, 's1': 1}, # 'policy1': 2
    (0, 1, 1, 2): {'policies': {'policy1': 8, 'policy2': 8}, 's1': 2},
    (0, 1, 1, 3): {'policies': {'policy1': 4, 'policy2': 8, 'policy3': 12}, 's1': 3},
    (0, 1, 2, 1): {'policies': {'policy1': 2, 'policy2': 4}, 's1': 4},
    (0, 1, 2, 2): {'policies': {'policy1': 8, 'policy2': 8}, 's1': 5},
    (0, 1, 2, 3): {'policies': {'policy1': 4, 'policy2': 8, 'policy3': 12}, 's1': 6},
    (0, 1, 3, 1): {'policies': {'policy1': 2, 'policy2': 4}, 's1': 7},
    (0, 1, 3, 2): {'policies': {'policy1': 8, 'policy2': 8}, 's1': 8},
    (0, 1, 3, 3): {'policies': {'policy1': 4, 'policy2': 8, 'policy3': 12}, 's1': 9}
}


def row(a, b):
    return a == b


params = [["policy_aggregation", result, expected_results, ['policies', 's1'], [row]]]


class GenericTest(make_generic_test(params)):
    pass


if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd

from cadCAD import configs
from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from testing.generic_test import make_generic_test

exec_mode = ExecutionMode()

first_config = configs
exec_ctx = ExecutionContext(context=exec_mode.local_mode)
run = Executor(exec_context=exec_ctx, configs=first_config)

raw_result, tensor_field, sessions = run.execute()
result = pd.DataFrame(raw_result)


def get_expected_results(run):
    return {
        (0, run, 0, 0): {
            'external_data': {'ds1': None, 'ds2': None, 'ds3': None},
            'increment': 0,
            'policies': {'ds1': None, 'ds2': None, 'ds3': None}
        },
        (0, run, 1, 1): {
            'external_data': {'ds1': 0, 'ds2': 0, 'ds3': 1},
             'increment': 1,
             'policies': {'ds1': 0, 'ds2': 0, 'ds3': 1}
        },
        (0, run, 1, 2): {
            'external_data': {'ds1': 1, 'ds2': 40, 'ds3': 5},
             'increment': 2,
             'policies': {'ds1': 1, 'ds2': 40, 'ds3': 5}
        },
        (0, run, 1, 3): {
            'external_data': {'ds1': 2, 'ds2': 40, 'ds3': 5},
             'increment': 3,
             'policies': {'ds1': 2, 'ds2': 40, 'ds3': 5}
        },
        (0, run, 2, 1): {
            'external_data': {'ds1': 3, 'ds2': 40, 'ds3': 5},
             'increment': 4,
             'policies': {'ds1': 3, 'ds2': 40, 'ds3': 5}
        },
        (0, run, 2, 2): {
            'external_data': {'ds1': 4, 'ds2': 40, 'ds3': 5},
             'increment': 5,
             'policies': {'ds1': 4, 'ds2': 40, 'ds3': 5}
        },
        (0, run, 2, 3): {
            'external_data': {'ds1': 5, 'ds2': 40, 'ds3': 5},
            'increment': 6,
            'policies': {'ds1': 5, 'ds2': 40, 'ds3': 5}
        },
        (0, run, 3, 1): {
            'external_data': {'ds1': 6, 'ds2': 40, 'ds3': 5},
            'increment': 7,
            'policies': {'ds1': 6, 'ds2': 40, 'ds3': 5}
        },
        (0, run, 3, 2): {
            'external_data': {'ds1': 7, 'ds2': 40, 'ds3': 5},
            'increment': 8,
            'policies': {'ds1': 7, 'ds2': 40, 'ds3': 5}
        },
        (0, run, 3, 3): {
            'external_data': {'ds1': 8, 'ds2': 40, 'ds3': 5},
            'increment': 9,
            'policies': {'ds1': 8, 'ds2': 40, 'ds3': 5}
        },
        (0, run, 4, 1): {
            'external_data': {'ds1': 9, 'ds2': 40, 'ds3': 5},
            'increment': 10,
            'policies': {'ds1': 9, 'ds2': 40, 'ds3': 5}
        },
        (0, run, 4, 2): {
            'external_data': {'ds1': 10, 'ds2': 40, 'ds3': 5},
            'increment': 11,
            'policies': {'ds1': 10, 'ds2': 40, 'ds3': 5}
        },
        (0, run, 4, 3): {
            'external_data': {'ds1': 11, 'ds2': 40, 'ds3': 5},
            'increment': 12,
            'policies': {'ds1': 11, 'ds2': 40, 'ds3': 5}
        }
    }


expected_results = {}
expected_results_1 = get_expected_results(1)
expected_results_2 = get_expected_results(2)
expected_results.update(expected_results_1)
expected_results.update(expected_results_2)


def row(a, b):
    return a == b


params = [["external_dataset", result, expected_results, ['increment', 'external_data', 'policies'], [row]]]


class GenericTest(make_generic_test(params)):
    pass


if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd

from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from cadCAD import configs
from cadCAD.utils.sys_exec import to_pandas_df

from testing.models import param_sweep
from testing.generic_test import make_generic_test
from testing.models.param_sweep import some_function, g as sweep_params

from distroduce.engine.execution import distributed_simulations, transform
from distroduce.session import sc_alt as sc

exec_mode = ExecutionMode()
distributed_sims = distributed_simulations(transform)
distributed_ctx = ExecutionContext(context=exec_mode.distributed, method=distributed_sims)
run = Executor(exec_context=distributed_ctx, configs=configs, spark_context=sc)

# sim, run, substep, timestep
def get_expected_results(subset, run, beta, gamma):
    return {
        (subset, run, 0, 0): {'policies': {}, 'sweeped': {}, 'alpha': 0, 'beta': 0},
        (subset, run, 1, 1): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 1, 2): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 1, 3): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 2, 1): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 2, 2): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 2, 3): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 3, 1): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 3, 2): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 3, 3): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 4, 1): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 4, 2): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 4, 3): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': {'beta': beta, 'gamma': gamma}, 'alpha': 1, 'beta': beta},
        (subset, run, 5, 1): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': beta, 'alpha': 1, 'beta': beta},
        (subset, run, 5, 2): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': beta, 'alpha': 1, 'beta': beta},
        (subset, run, 5, 3): {'policies': {'gamma': gamma, 'omega': 7}, 'sweeped': beta, 'alpha': 1, 'beta': beta}
    }


def generate_expected(sweep_params):
    def template(sweep_params):
        subset_count = max(len(x) for x in list(sweep_params.values()))
        expected_results, expected_results_1, expected_results_2 = {}, {}, {}
        for subset in range(subset_count):
            expected_results_1a = get_expected_results(subset, 1, 2, 3)
            expected_results_1b = get_expected_results(subset, 2, 2, 3)
            expected_results_1.update(expected_results_1a)
            expected_results_1.update(expected_results_1b)

            expected_results_2 = {}
            expected_results_2a = get_expected_results(subset, 1, some_function, 4)
            expected_results_2b = get_expected_results(subset, 2, some_function, 4)
            expected_results_2.update(expected_results_2a)
            expected_results_2.update(expected_results_2b)

            expected_results.update(expected_results_1)
            expected_results.update(expected_results_2)

            yield expected_results

    merged_expected = list(template(sweep_params))
    result = {}
    for d in merged_expected:
        result.update(d)

    return result


def row(a, b):
    return a == b


def create_test_params(feature, fields):
    raw_result, tensor_fields, sessions = run.execute()
    df: pd.DataFrame = to_pandas_df(raw_result)
    expected = generate_expected(sweep_params)
    return [[feature, df, expected, fields, [row]]]


params = list(create_test_params("param_sweep", ['alpha', 'beta', 'policies', 'sweeped']))


class GenericTest(make_generic_test(params)):
    pass


if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
from datetime import datetime
from functools import reduce

add = lambda a, b: a + b


'''
Function that maintains the state of users in a chat room
'''
def update_users(users, actions, action_types=['send','enter','exit']):
    users = deepcopy(users)
    for action_type in action_types:
        if action_type in actions['types']:
            for msg in actions['messages']:
                if msg['action'] == 'send' and action_type == 'send':
                    continue
                elif msg['action'] == 'enter' and action_type == 'enter':
                    for user in msg['sender']:
                        users.append(user) # register_entered
                elif msg['action'] == 'exit' and action_type == 'exit':
                    for user in msg['sender']:
                        users.remove(user) # remove_exited
    return users


'''
State Update used to counts sent messages
'''
def count_messages(_g, step, sL, s, actions, **kwargs):
    return 'total_msg_count', s['total_msg_count'] + reduce(add, actions['msg_counts'])


'''
State Update used to create to log message events
'''
def send_message(state):
    return lambda _g, step, sL, s, actions, **kwargs: (
        state,
        {
            'users': update_users(s[state]['users'], actions),
            'messages': actions['messages'],
            'msg_counts': reduce(add, actions['msg_counts']),
            'send_times': reduce(add, actions['send_times'])
        }
    )


'''
State Update used to sum the time taken for the Kafka Producer to send messages between users
'''
def add_send_time(_g, step, sL, s, actions, **kwargs):
    return 'total_send_time', s['total_send_time'] + reduce(add, actions['send_times'])


'''
State Update used to record the event record creation time
'''
def current_time(state):
    return lambda _g, step, sL, s, actions, **kwargs: (state, datetime.now())


import pandas as pd
from datetime import timedelta
import pprint as pp

from cadCAD.utils import SilentDF
from cadCAD.configuration.utils import time_step, config_sim, var_trigger, var_substep_trigger, env_trigger, psub_list
from cadCAD.configuration.utils.userDefinedObject import udoPipe, UDO
from simulations.regression_tests.experiments import udo1_exp

DF = SilentDF(pd.read_csv('simulations/external_data/output.csv'))


class udoExample(object):
    def __init__(self, x, dataset=None):
        self.x = x
        self.mem_id = str(hex(id(self)))
        self.ds = dataset # for setting ds initially or querying
        self.perception = {}

    def anon(self, f):
        return f(self)

    def updateX(self):
        self.x += 1
        return self

    def updateDS(self):
        self.ds.iloc[0,0] -= 10
        # pp.pprint(self.ds)
        return self

    def perceive(self, s):
        self.perception = self.ds[
            (self.ds['run'] == s['run']) & (self.ds['substep'] == s['substep']) & (self.ds['timestep'] == s['timestep'])
        ].drop(columns=['run', 'substep']).to_dict()
        return self

    def read(self, ds_uri):
        self.ds = SilentDF(pd.read_csv(ds_uri))
        return self

    def write(self, ds_uri):
        pd.to_csv(ds_uri)

    pass


state_udo = UDO(udo=udoExample(0, DF), masked_members=['obj', 'perception'])
policy_udoA = UDO(udo=udoExample(0, DF), masked_members=['obj', 'perception'])
policy_udoB = UDO(udo=udoExample(0, DF), masked_members=['obj', 'perception'])


sim_config = config_sim({
    "N": 3,
    "T": range(4)
})

state_dict = {
    'increment': 0,
    'state_udo': state_udo, 'state_udo_tracker': 0,
    'state_udo_perception_tracker': {"ds1": None, "ds2": None, "ds3": None, "timestep": None},
    'udo_policies': {'udo_A': policy_udoA, 'udo_B': policy_udoB},
    'udo_policy_tracker': (0, 0),
    'timestamp': '2019-01-01 00:00:00'
}

psu_steps = ['m1', 'm2', 'm3']
system_substeps = len(psu_steps)
var_timestep_trigger = var_substep_trigger([0, system_substeps])
env_timestep_trigger = env_trigger(system_substeps)
psu_block = {k: {"policies": {}, "variables": {}} for k in psu_steps}

def udo_policyA(_g, step, sL, s, **kwargs):
    s['udo_policies']['udo_A'].updateX()
    return {'udo_A': udoPipe(s['udo_policies']['udo_A'])}
# policies['a'] = udo_policyA
for m in psu_steps:
    psu_block[m]['policies']['a'] = udo_policyA

def udo_policyB(_g, step, sL, s, **kwargs):
    s['udo_policies']['udo_B'].updateX()
    return {'udo_B': udoPipe(s['udo_policies']['udo_B'])}
# policies['b'] = udo_policyB
for m in psu_steps:
    psu_block[m]['policies']['b'] = udo_policyB


# policies = {"p1": udo_policyA, "p2": udo_policyB}
# policies = {"A": udo_policyA, "B": udo_policyB}

def add(y: str, added_val):
    return lambda _g, step, sL, s, _input, **kwargs: (y, s[y] + added_val)
# state_updates['increment'] = add('increment', 1)
for m in psu_steps:
    psu_block[m]["variables"]['increment'] = add('increment', 1)


# @curried
def perceive(s, self):
    self.perception = self.ds[
        (self.ds['run'] == s['run']) & (self.ds['substep'] == s['substep']) & (self.ds['timestep'] == s['timestep'])
    ].drop(columns=['run', 'substep']).to_dict()
    return self


def state_udo_update(_g, step, sL, s, _input, **kwargs):
    y = 'state_udo'
    # s['hydra_state'].updateX().anon(perceive(s))
    s['state_udo'].updateX().perceive(s).updateDS()
    x = udoPipe(s['state_udo'])
    return y, x
for m in psu_steps:
    psu_block[m]["variables"]['state_udo'] = state_udo_update


def track(destination, source):
    return lambda _g, step, sL, s, _input, **kwargs: (destination, s[source].x)
state_udo_tracker = track('state_udo_tracker', 'state_udo')
for m in psu_steps:
    psu_block[m]["variables"]['state_udo_tracker'] = state_udo_tracker


def track_state_udo_perception(destination, source):
    def id(past_perception):
        if len(past_perception) == 0:
            return state_dict['state_udo_perception_tracker']
        else:
            return past_perception
    return lambda _g, step, sL, s, _input, **kwargs: (destination, id(s[source].perception))
state_udo_perception_tracker = track_state_udo_perception('state_udo_perception_tracker', 'state_udo')
for m in psu_steps:
    psu_block[m]["variables"]['state_udo_perception_tracker'] = state_udo_perception_tracker


def view_udo_policy(_g, step, sL, s, _input, **kwargs):
    return 'udo_policies', _input
for m in psu_steps:
    psu_block[m]["variables"]['udo_policies'] = view_udo_policy


def track_udo_policy(destination, source):
    def val_switch(v):
        if isinstance(v, pd.DataFrame) is True or isinstance(v, SilentDF) is True:
            return SilentDF(v)
        else:
            return v.x
    return lambda _g, step, sL, s, _input, **kwargs: (destination, tuple(val_switch(v) for _, v in s[source].items()))
udo_policy_tracker = track_udo_policy('udo_policy_tracker', 'udo_policies')
for m in psu_steps:
    psu_block[m]["variables"]['udo_policy_tracker'] = udo_policy_tracker


def update_timestamp(_g, step, sL, s, _input, **kwargs):
    y = 'timestamp'
    return y, time_step(dt_str=s[y], dt_format='%Y-%m-%d %H:%M:%S', _timedelta=timedelta(days=0, minutes=0, seconds=1))
for m in psu_steps:
    psu_block[m]["variables"]['timestamp'] = var_timestep_trigger(y='timestamp', f=update_timestamp)
    # psu_block[m]["variables"]['timestamp'] = var_trigger(
    #     y='timestamp', f=update_timestamp,
    #     pre_conditions={'substep': [0, system_substeps]}, cond_op=lambda a, b: a and b
    # )
    # psu_block[m]["variables"]['timestamp'] = update_timestamp

# New Convention
partial_state_update_blocks = psub_list(psu_block, psu_steps)
udo1_exp.append_configs(
    sim_configs=sim_config,
    initial_state=state_dict,
    partial_state_update_blocks=partial_state_update_blocks
)

print()
print("State Updates:")
pp.pprint(partial_state_update_blocks)
print()


from datetime import datetime

# ToDo: Model Async communication here
# ToDo: Configuration of Input
def configure_producer(msg, config):
    from kafka import KafkaProducer
    def send_messages(events):
        producer = KafkaProducer(**config)

        start_timestamp = datetime.now()
        for event in range(events):
            producer.send('test', msg(event)).get()
        delta = datetime.now() - start_timestamp

        return start_timestamp, delta.total_seconds()

    return send_messages



import pprint
from typing import Dict, List

import pandas as pd
from tabulate import tabulate

from cadCAD.configuration.utils import env_trigger, var_substep_trigger, config_sim, psub_list
from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from cadCAD.configuration import Experiment
from cadCAD import configs

pp = pprint.PrettyPrinter(indent=4)


def some_function(x):
    return x


g: Dict[str, List[int]] = {
    'alpha': [1],
    'beta': [2, 5],
    'gamma': [3, 4],
    'omega': [some_function]
}

psu_steps = ['1', '2', '3']
system_substeps = len(psu_steps)
var_timestep_trigger = var_substep_trigger([0, system_substeps])
env_timestep_trigger = env_trigger(system_substeps)
env_process = {}


# Policies
def gamma(_params, step, sH, s):
    return {'gamma': _params['gamma']}


def omega(_params, step, sH, s):
    return {'omega': _params['omega'](7)}


# Internal States
def alpha(_params, step, sH, s, _input):
    return 'alpha', _params['alpha']

def alpha_plus_gamma(_params, step, sH, s, _input):
    return 'alpha_plus_gamma', _params['alpha'] + _params['gamma']


def beta(_params, step, sH, s, _input):
    return 'beta', _params['beta']


def policies(_params, step, sH, s, _input):
    return 'policies', _input


def sweeped(_params, step, sH, s, _input):
    return 'sweeped', {'beta': _params['beta'], 'gamma': _params['gamma']}


genesis_states = {
    'alpha_plus_gamma': 0,
    'alpha': 0,
    'beta': 0,
    'policies': {},
    'sweeped': {}
}

env_process['sweeped'] = env_timestep_trigger(trigger_field='timestep', trigger_vals=[5], funct_list=[lambda _g, x: _g['beta']])

sim_config = config_sim(
    {
        "N": 2,
        "T": range(5),
        "M": g,
    }
)

psu_block = {k: {"policies": {}, "variables": {}} for k in psu_steps}
for m in psu_steps:
    psu_block[m]['policies']['gamma'] = gamma
    psu_block[m]['policies']['omega'] = omega
    psu_block[m]["variables"]['alpha'] = alpha_plus_gamma
    psu_block[m]["variables"]['alpha_plus_gamma'] = alpha
    psu_block[m]["variables"]['beta'] = beta
    psu_block[m]['variables']['policies'] = policies
    psu_block[m]["variables"]['sweeped'] = var_timestep_trigger(y='sweeped', f=sweeped)

psubs = psub_list(psu_block, psu_steps)
print()
pp.pprint(psu_block)
print()

exp = Experiment()
exp.append_configs(
    sim_configs=sim_config,
    initial_state=genesis_states,
    env_processes=env_process,
    partial_state_update_blocks=psubs
)

exec_mode = ExecutionMode()
local_proc_ctx = ExecutionContext(context=exec_mode.local_mode)
run = Executor(exec_context=local_proc_ctx, configs=configs)

raw_result, tensor_field, sessions = run.execute()
result = pd.DataFrame(raw_result)
print()
print("Tensor Field:")
print(tabulate(tensor_field, headers='keys', tablefmt='psql'))
print("Output:")
print(tabulate(result, headers='keys', tablefmt='psql'))
print()


from copy import deepcopy


def sanitize_config(config):
    for key, value in config.kwargs.items():
        if key == 'state_dict':
            config.initial_state = value
        elif key == 'seed':
            config.seeds = value
        elif key == 'mechanisms':
            config.partial_state_updates = value

    if config.initial_state == {}:
        raise Exception('The initial conditions of the system have not been set')


def sanitize_partial_state_updates(partial_state_updates):
    new_partial_state_updates = deepcopy(partial_state_updates)
    def rename_keys(d):
        if 'behaviors' in d:
            d['policies'] = d.pop('behaviors')

        if 'states' in d:
            d['variables'] = d.pop('states')

    if isinstance(new_partial_state_updates, list):
        for v in new_partial_state_updates:
            rename_keys(v)
    elif isinstance(new_partial_state_updates, dict):
        for k, v in new_partial_state_updates.items():
            rename_keys(v)

    del partial_state_updates
    return new_partial_state_updates



from funcy import curry
from cadCAD.configuration.utils import ep_time_step, time_step


def increment(y, incr_by):
    return lambda _g, step, sL, s, _input, **kwargs: (y, s[y] + incr_by)


def track(y):
    return lambda _g, step, sL, s, _input, **kwargs: (y, s[y].x)


def simple_state_update(y, x):
    return lambda _g, step, sH, s, _input, **kwargs: (y, x)


def simple_policy_update(y):
    return lambda _g, step, sH, s, **kwargs: y


def update_timestamp(y, timedelta, format):
    return lambda _g, step, sL, s, _input, **kwargs: (
        y,
        ep_time_step(s, dt_str=s[y], fromat_str=format, _timedelta=timedelta)
    )


def apply(f, y: str, incr_by: int):
    return lambda _g, step, sL, s, _input, **kwargs: (y, curry(f)(s[y])(incr_by))


def add(y: str, incr_by):
    return apply(lambda a, b: a + b, y, incr_by)


def increment_state_by_int(y: str, incr_by: int):
    return lambda _g, step, sL, s, _input, **kwargs: (y, s[y] + incr_by)


def s(y, x):
    return lambda _g, step, sH, s, _input, **kwargs: (y, x)


def time_model(y, substeps, time_delta, ts_format='%Y-%m-%d %H:%M:%S'):
    def apply_incriment_condition(s):
        if s['substep'] == 0 or s['substep'] == substeps:
            return y, time_step(dt_str=s[y], dt_format=ts_format, _timedelta=time_delta)
        else:
            return y, s[y]
    return lambda _g, step, sL, s, _input, **kwargs: apply_incriment_condition(s)



from datetime import datetime

from cadCAD import configs
from cadCAD.configuration.utils import config_sim
from cadCAD.engine import ExecutionMode, ExecutionContext, Executor

from simulations.regression_tests.poc.simulation import main, sim_composition

from distroduce.engine.execution import transform, distributed_simulations
from distroduce.session import sc_alt as sc

if __name__ == "__main__":
    # Initial States
    initial_conditions = {
        'record_creation': datetime.now(),
        'client_a': {'users': [], 'messages': [], 'msg_count': 0, 'send_time': 0.0},
        'client_b': {'users': [], 'messages': [], 'msg_count': 0, 'send_time': 0.0},
        'total_msg_count': 0,
        'total_send_time': 0.000000
    }

    '''
        Simulation Configuration:
        N = Simulation Runs
        T = Timesteps for each Partial State Update Block
    '''

    sim_config_dict = {
        "N":  3,
        "T": range(5),
    }
    # sim_config_dict = {
    #     "N": 1000,
    #     "T": range(5000),
    # }
    sim_config = config_sim(sim_config_dict)

    bootstrap_servers = ['localhost:9092']
    # bootstrap_servers = ['10.0.0.9:9092']

    # Configuration for Kafka Producer
    kafkaConfig = {
        'send_topic': 'test',
        'producer_config': {
            'bootstrap_servers': bootstrap_servers,
            'acks': 'all'
        }
    }

    parts = sim_config_dict['N']
    exec_mode = ExecutionMode()

    distributed_sims = distributed_simulations(transform(publish=False))
    distributed_ctx = ExecutionContext(context=exec_mode.distributed, method=distributed_sims, additional_objs=kafkaConfig)
    run = Executor(exec_context=distributed_ctx, configs=configs, spark_context=sc)

    main(run, sim_config, initial_conditions, sim_composition)

    # make default behind scenes
    # def rdd_trans_action(dist_obj_config, func_params_kv, rdd):
    #     # dist_obj_config: Object Injection in a good way
    #     # Sink for comparison / access
    #     simulate = dist_simulation(dist_obj_config, func_params_kv)
    #     spark_trans_action = rdd.map(simulate).collect()
    #     return list(spark_trans_action)

    # 10 executors: 10 - m4.2xlarge
    # spark.dynamicAllocation.enabled - true

    # new spark version
    # 10 M in 1.1 / 4.0 min - application_1576607651341_0002
    # .config("spark.executor.memory", "20g") \
    # spark.executor.memory - 13.3GB (total 147GB [executors included])
    # spark.executor.cores - 16 (total 16 x 10 = 160)

    # new spark version
    # 12 M in 1.6 / 4.9 min - application_1576693884747_0002
    # .config("spark.executor.memory", "25g") \
    # spark.executor.memory - 13.3GB (total 147GB [executors included])
    # spark.executor.cores - 16 (total 16 x 10 = 160)

    # new spark version
    # 14 M in 1.6 / 5.3 min - application_1576696344918_0002
    # .config("spark.executor.memory", "30g") \
    # spark.executor.memory - 13.3GB (total 147GB [executors included])
    # spark.executor.cores - 16 (total 16 x 10 = 160)

    # new spark version
    # 16 M in 1.9 / 6.3 min - application_1576696344918_0003
    # .config("spark.executor.memory", "30g") \
    # spark.executor.memory - 13.3GB (total 147GB [executors included])
    # spark.executor.cores - 16 (total 16 x 10 = 160)

    # new spark version
    # 18 M in 2 / 6.8 min - application_1576696344918_0004
    # .config("spark.executor.memory", "30g") \
    # spark.executor.memory - 13.3GB (total 147GB [executors included])
    # spark.executor.cores - 16 (total 16 x 10 = 160)

    # new spark version
    # 20 M in 2.1 / 7.6 min - application_1576696344918_0005
    # .config("spark.executor.memory", "30g") \
    # spark.executor.memory - 13.3GB (total 147GB [executors included])
    # spark.executor.cores - 16 (total 16 x 10 = 160)
    # ---------------------------------------------------------------------
    # 10 M in 2.1 / 4.1 min - application_???
    # .config("spark.executor.memory", "20g") \
    # spark.executor.memory - 13.5GB
    # spark.executor.cores - 16

    # 10 M in 2.1 / 5.3 min - application_1574186919629_0002
    # .config("spark.executor.memory", "20g") \
    # spark.executor.memory - 13.3GB
    # spark.executor.cores - 16



    # 2 M in 2.0 / 4.9 min - application_1574177687848_0008
    # .config("spark.executor.memory", "20g") \
    # spark.executor.memory - 12.7 GB
    # spark.executor.cores - 16
    # 2 M in 1.8 / 4.9 min - application_1574177687848_0006
    # spark.executor.memory - 13.3GB
    # spark.executor.cores - 16

    # 169 executors: 10 - m4.2xlarge
    # spark.dynamicAllocation.enabled - true
    # spark.executor.memory - 435MB
    # spark.executor.cores - 1

    # 8 M in 2.3 / 4.9 min - application_1574177687848_0004
    #     .config("spark.executor.memory", "2g") \
    #     .config("spark.executor.cores", "2") \
    # 8 M in 2.3 / 4.9 min - application_1574177687848_0003
    #     .config("spark.executor.memory", "2g") \
    #     .config("spark.executor.cores", "1") \
    # 8 M in 2.4 / 4.9 min - application_1574177687848_0002
    #     .config("spark.executor.memory", "1g") \
    #     .config("spark.executor.cores", "1") \

    # 34 executors: 5 - m4.2xlarge
    # 8 M in 3.7 / 6.3 min - application_1574122229644_0008
    #     .config("spark.executor.memory", "1g") \
    #     .config("spark.executor.cores", "1") \
    # 8 M in 3.6 / 6.1 min - application_1574122229644_0008
    #     .config("spark.executor.memory", "2g") \
    #     .config("spark.executor.cores", "1") \
    # 8 M in 3.7 / 6.2 min - application_1574122229644_0003
    #     .config("spark.executor.memory", "5g") \
    #     .config("spark.executor.cores", "5") \
    # 8 M in 3.7 / 6.2 min - application_1574122229644_0003
    #     .config("spark.executor.memory", "3g") \
    #     .config("spark.executor.cores", "4") \
    # 8 M in 3.7 / 6.3 min - application_1574122229644_0005
    #     .config("spark.executor.memory", "2g") \
    #     .config("spark.executor.cores", "5") \
    # 6 M in 2.9 / 4.8 min - application_1574122229644_0002
    #     .config("spark.executor.memory", "3g") \
    #     .config("spark.executor.cores", "4") \

    # ? executors: 5 - m4.xlarge
    # 6 M in 4.1 / 5.7 min - application_1574117954818_0003

    # 169 executors: 10 - m4.2xlarge
    # spark.dynamicAllocation.enabled - true



import pandas as pd
import pprint as pp
# from fn.func import curried
from datetime import timedelta

from cadCAD.utils import SilentDF #, val_switch
from cadCAD.configuration.utils import time_step, config_sim
from cadCAD.configuration.utils.userDefinedObject import udoPipe, UDO
from simulations.regression_tests.experiments import udo1_exp

DF = SilentDF(pd.read_csv('simulations/external_data/output.csv'))

class udoExample(object):
    def __init__(self, x, dataset=None):
        self.x = x
        self.mem_id = str(hex(id(self)))
        self.ds = dataset # for setting ds initially or querying
        self.perception = {}

    def anon(self, f):
        return f(self)

    def updateX(self):
        self.x += 1
        return self

    def perceive(self, s):
        self.perception = self.ds[
            (self.ds['run'] == s['run']) & (self.ds['substep'] == s['substep']) & (self.ds['timestep'] == s['timestep'])
        ].drop(columns=['run', 'substep']).to_dict()
        return self

    def read(self, ds_uri):
        self.ds = SilentDF(pd.read_csv(ds_uri))
        return self

    def write(self, ds_uri):
        pd.to_csv(ds_uri)

    # Generic update function

    pass

# can be accessed after an update within the same substep and timestep
state_udo = UDO(udo=udoExample(0, DF), masked_members=['obj', 'perception'])
policy_udoA = UDO(udo=udoExample(0, DF), masked_members=['obj', 'perception'])
policy_udoB = UDO(udo=udoExample(0, DF), masked_members=['obj', 'perception'])

def udo_policyA(_g, step, sL, s, **kwargs):
    s['udo_policies']['udo_A'].updateX()
    return {'udo_A': udoPipe(s['udo_policies']['udo_A'])}

def udo_policyB(_g, step, sL, s, **kwargs):
    s['udo_policies']['udo_B'].updateX()
    return {'udo_B': udoPipe(s['udo_policies']['udo_B'])}


policies = {"p1": udo_policyA, "p2": udo_policyB}

state_dict = {
    'increment': 0,
    'state_udo': state_udo, 'state_udo_tracker_a': 0, 'state_udo_tracker_b': 0,
    'state_udo_perception_tracker': {"ds1": None, "ds2": None, "ds3": None, "timestep": None},
    'udo_policies': {'udo_A': policy_udoA, 'udo_B': policy_udoB},
    'udo_policy_tracker_a': (0, 0), 'udo_policy_tracker_b': (0, 0),
    'timestamp': '2019-01-01 00:00:00'
}

# @curried
def perceive(s, self):
    self.perception = self.ds[
        (self.ds['run'] == s['run']) & (self.ds['substep'] == s['substep']) & (self.ds['timestep'] == s['timestep'])
    ].drop(columns=['run', 'substep']).to_dict()
    return self

def view_udo_policy(_g, step, sL, s, _input, **kwargs):
    return 'udo_policies', _input

def state_udo_update(_g, step, sL, s, _input, **kwargs):
    y = 'state_udo'
    # s['hydra_state'].updateX().anon(perceive(s))
    s['state_udo'].updateX().perceive(s)
    x = udoPipe(s['state_udo'])
    return y, x

def increment(y, incr_by):
    return lambda _g, step, sL, s, _input, **kwargs: (y, s[y] + incr_by)

def track(destination, source):
    return lambda _g, step, sL, s, _input, **kwargs: (destination, s[source].x)

def track_udo_policy(destination, source):
    def val_switch(v):
        if isinstance(v, pd.DataFrame) is True or isinstance(v, SilentDF) is True:
            return SilentDF(v)
        else:
            return v.x
    return lambda _g, step, sL, s, _input, **kwargs: (destination, tuple(val_switch(v) for _, v in s[source].items()))

def track_state_udo_perception(destination, source):
    def id(past_perception):
        if len(past_perception) == 0:
            return state_dict['state_udo_perception_tracker']
        else:
            return past_perception
    return lambda _g, step, sL, s, _input, **kwargs: (destination, id(s[source].perception))


def time_model(y, substeps, time_delta, ts_format='%Y-%m-%d %H:%M:%S'):
    def apply_incriment_condition(s):
        if s['substep'] == 0 or s['substep'] == substeps:
            return y, time_step(dt_str=s[y], dt_format=ts_format, _timedelta=time_delta)
        else:
            return y, s[y]
    return lambda _g, step, sL, s, _input, **kwargs: apply_incriment_condition(s)


states = {
    'increment': increment('increment', 1),
    'state_udo_tracker_a': track('state_udo_tracker_a', 'state_udo'),
    'state_udo': state_udo_update,
    'state_udo_perception_tracker': track_state_udo_perception('state_udo_perception_tracker', 'state_udo'),
    'state_udo_tracker_b': track('state_udo_tracker_b', 'state_udo'),
    'udo_policy_tracker_a': track_udo_policy('udo_policy_tracker_a', 'udo_policies'),
    'udo_policies': view_udo_policy,
    'udo_policy_tracker_b': track_udo_policy('udo_policy_tracker_b', 'udo_policies')
}

substeps=3
update_timestamp = time_model(
    'timestamp',
    substeps=3,
    time_delta=timedelta(days=0, minutes=0, seconds=1),
    ts_format='%Y-%m-%d %H:%M:%S'
)
states['timestamp'] = update_timestamp

PSUB = {
    'policies': policies,
    'states': states
}

# needs M1&2 need behaviors
partial_state_update_blocks = [PSUB] * substeps
# pp.pprint(partial_state_update_blocks)

sim_config = config_sim({
    "N": 2,
    "T": range(4)
})

udo1_exp.append_configs(
    sim_configs=sim_config,
    initial_state=state_dict,
    seeds={},
    raw_exogenous_states={},
    env_processes={},
    partial_state_update_blocks=partial_state_update_blocks,
    # policy_ops=[lambda a, b: {**a, **b}]
)

print()
print("State Updates:")
pp.pprint(partial_state_update_blocks)
print()


from pprint import pprint
import pandas as pd
from tabulate import tabulate

from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from simulations.regression_tests.models import config2
from cadCAD import configs

exec_mode = ExecutionMode()

local_proc_ctx = ExecutionContext(context=exec_mode.local_mode)
run = Executor(exec_context=local_proc_ctx, configs=configs)

raw_result, tensor_fields, sessions = run.execute()
result = pd.DataFrame(raw_result)
print(tabulate(tensor_fields[0], headers='keys', tablefmt='psql'))
pprint(sessions)
print(tabulate(result, headers='keys', tablefmt='psql'))



import pandas as pd
from tabulate import tabulate
# The following imports NEED to be in the exact order
from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from documentation.examples import sys_model_B
from cadCAD import configs

exec_mode = ExecutionMode()

print("Simulation Execution: Single Configuration")
print()
local_proc_ctx = ExecutionContext(context=exec_mode.local_mode)
run = Executor(exec_context=local_proc_ctx, configs=configs)

raw_result, tensor_field, sessions = run.execute()
result = pd.DataFrame(raw_result)
print()
print("Tensor Field: sys_model_B")
print(tabulate(tensor_field, headers='keys', tablefmt='psql'))
print("Output:")
print(tabulate(result, headers='keys', tablefmt='psql'))
print()



from pprint import pprint
from typing import Any, Callable, Dict, List, Tuple
from copy import deepcopy
from functools import reduce
from funcy import curry

from cadCAD.utils import flatten
from cadCAD.engine.utils import engine_exception

id_exception: Callable = curry(engine_exception)(KeyError)(KeyError)(None)


class Executor:
    def __init__(
        self,
        policy_ops,
        policy_update_exception: Callable = id_exception,
        state_update_exception: Callable = id_exception
    ) -> None:

        self.policy_ops = policy_ops
        self.state_update_exception = state_update_exception
        self.policy_update_exception = policy_update_exception

    def get_policy_input(
        self,
        sweep_dict: Dict[str, List[Any]],
        sub_step: int,
        sL: List[Dict[str, Any]],
        s: Dict[str, Any],
        funcs: List[Callable],
        additional_objs
    ) -> Dict[str, Any]:

        ops = self.policy_ops

        def get_col_results(sweep_dict, sub_step, sL, s, funcs):
            # pprint(sweep_dict)
            # exit()
            def policy_scope_tuner(additional_objs, f):
                if additional_objs is None:
                    return f(sweep_dict, sub_step, sL, s)
                else:
                    return f(sweep_dict, sub_step, sL, s, additional_objs)
            return list(map(lambda f: policy_scope_tuner(additional_objs, f), funcs))

        def compose(init_reduction_funct, funct_list, val_list):
            result, i = None, 0
            composition = lambda x: [reduce(init_reduction_funct, x)] + funct_list
            for g in composition(val_list):
                if i == 0:
                    result = g
                    i = 1
                else:
                    result = g(result)
            return result

        col_results = get_col_results(sweep_dict, sub_step, sL, s, funcs)
        key_set = list(set(list(reduce(lambda a, b: a + b, list(map(lambda x: list(x.keys()), col_results))))))
        new_dict = {k: [] for k in key_set}
        for d in col_results:
            for k in d.keys():
                new_dict[k].append(d[k])

        ops_head, *ops_tail = ops
        return {
            k: compose(
                init_reduction_funct=ops_head,
                funct_list=ops_tail,
                val_list=val_list
            ) for k, val_list in new_dict.items()
        }

    def apply_env_proc(
        self,
        sweep_dict,
        env_processes: Dict[str, Callable],
        state_dict: Dict[str, Any]
    ) -> Dict[str, Any]:

        def env_composition(target_field, state_dict, target_value):
            function_type = type(lambda x: x)
            env_update = env_processes[target_field]
            if isinstance(env_update, list):
                for f in env_update:
                    target_value = f(sweep_dict, target_value)
            elif isinstance(env_update, function_type):
                target_value = env_update(state_dict, sweep_dict, target_value)
            else:
                target_value = env_update

            return target_value

        filtered_state_dict = {k: v for k, v in state_dict.items() if k in env_processes.keys()}
        env_proc_dict = {
            target_field: env_composition(target_field, state_dict, target_value)
            for target_field, target_value in filtered_state_dict.items()
        }

        for k, v in env_proc_dict.items():
            state_dict[k] = v

        return state_dict

    # mech_step
    def partial_state_update(
        self,
        sweep_dict: Dict[str, List[Any]],
        sub_step: int,
        sL,
        sH,
        state_funcs: List[Callable],
        policy_funcs: List[Callable],
        env_processes: Dict[str, Callable],
        time_step: int,
        run: int,
        additional_objs
    ) -> List[Dict[str, Any]]:

        # last_in_obj: Dict[str, Any] = MappingProxyType(sL[-1])
        last_in_obj: Dict[str, Any] = deepcopy(sL[-1])
        _input: Dict[str, Any] = self.policy_update_exception(
            self.get_policy_input(sweep_dict, sub_step, sH, last_in_obj, policy_funcs, additional_objs)
        )

        def generate_record(state_funcs):
            def state_scope_tuner(f):
                lenf = f.__code__.co_argcount
                if lenf == 5:
                    return self.state_update_exception(f(sweep_dict, sub_step, sH, last_in_obj, _input))
                elif lenf == 6:
                    return self.state_update_exception(f(sweep_dict, sub_step, sH, last_in_obj, _input, additional_objs))
            for f in state_funcs:
                yield state_scope_tuner(f)

        def transfer_missing_fields(source, destination):
            for k in source:
                if k not in destination:
                    destination[k] = source[k]
            del source
            return destination

        last_in_copy: Dict[str, Any] = transfer_missing_fields(last_in_obj, dict(generate_record(state_funcs)))
        last_in_copy: Dict[str, Any] = self.apply_env_proc(sweep_dict, env_processes, last_in_copy)
        last_in_copy['substep'], last_in_copy['timestep'], last_in_copy['run'] = sub_step, time_step, run

        sL.append(last_in_copy)
        del last_in_copy

        return sL

    # mech_pipeline - state_update_block
    def state_update_pipeline(
        self,
        sweep_dict: Dict[str, List[Any]],
        simulation_list,
        configs: List[Tuple[List[Callable], List[Callable]]],
        env_processes: Dict[str, Callable],
        time_step: int,
        run: int,
        additional_objs
    ) -> List[Dict[str, Any]]:

        sub_step = 0
        states_list_copy: List[Dict[str, Any]] = tuple(simulation_list[-1])
        genesis_states: Dict[str, Any] = states_list_copy[-1].copy()
#         genesis_states: Dict[str, Any] = states_list_copy[-1]

        if len(states_list_copy) == 1:
            genesis_states['substep'] = sub_step

        del states_list_copy
        states_list: List[Dict[str, Any]] = [genesis_states]

        sub_step += 1
        for [s_conf, p_conf] in configs:
            states_list: List[Dict[str, Any]] = self.partial_state_update(
                sweep_dict, sub_step, states_list, simulation_list, s_conf, p_conf, env_processes, time_step, run,
                additional_objs
            )
            sub_step += 1

        time_step += 1

        return states_list

    # state_update_pipeline
    def run_pipeline(
        self,
        sweep_dict: Dict[str, List[Any]],
        states_list: List[Dict[str, Any]],
        configs: List[Tuple[List[Callable], List[Callable]]],
        env_processes: Dict[str, Callable],
        time_seq: range,
        run: int,
        additional_objs
    ) -> List[List[Dict[str, Any]]]:
        time_seq: List[int] = [x + 1 for x in time_seq]
        simulation_list: List[List[Dict[str, Any]]] = [states_list]

        for time_step in time_seq:
            pipe_run: List[Dict[str, Any]] = self.state_update_pipeline(
                sweep_dict, simulation_list, configs, env_processes, time_step, run, additional_objs
            )

            _, *pipe_run = pipe_run
            simulation_list.append(pipe_run)

        return simulation_list

    def simulation(
        self,
        sweep_dict: Dict[str, List[Any]],
        states_list: List[Dict[str, Any]],
        configs,
        env_processes: Dict[str, Callable],
        time_seq: range,
        simulation_id: int,
        run: int,
        subset_id,
        subset_window,
        configured_N,
        # remote_ind
        additional_objs=None
    ):
        run += 1

        subset_window.appendleft(subset_id)
        latest_subset_id, previous_subset_id = tuple(subset_window)

        if configured_N == 1 and latest_subset_id > previous_subset_id:
            run -= 1

        def execute_run(sweep_dict, states_list, configs, env_processes, time_seq, _run) -> List[Dict[str, Any]]:
            def generate_init_sys_metrics(genesis_states_list, sim_id, _subset_id, _run, _subset_window):
                for D in genesis_states_list:
                    d = deepcopy(D)
                    d['simulation'], d['subset'], d['run'], d['substep'], d['timestep'] = \
                        sim_id, _subset_id, _run, 0, 0
                    yield d

            states_list_copy: List[Dict[str, Any]] = list(
                generate_init_sys_metrics(tuple(states_list), simulation_id, subset_id, run, subset_window)
            )

            first_timestep_per_run: List[Dict[str, Any]] = self.run_pipeline(
                sweep_dict, states_list_copy, configs, env_processes, time_seq, run, additional_objs
            )
            del states_list_copy

            return first_timestep_per_run

        pipe_run = flatten(
            [execute_run(sweep_dict, states_list, configs, env_processes, time_seq, run)]
        )

        return pipe_run



from cadCAD.configuration import Experiment

config1_exp = Experiment()
config2_exp = Experiment()
ext_ds_exp = Experiment()
hist_exp = Experiment()
policy_exp = Experiment()
sweep_exp = Experiment()
udo1_exp = Experiment()
udo2_exp = Experiment()



def get_home_dir(user):
    return f"s3://jupyterbackups/jupyter/{user}/"

def set_write_path(_spark_context, user, datafolder_path):
    return get_home_dir(user) + datafolder_path +f'_{_spark_context.applicationId}'



from pprint import pprint
from typing import Callable, Dict, List, Any, Tuple
from pathos.multiprocessing import ThreadPool as TPool
from pathos.multiprocessing import ProcessPool as PPool
from collections import Counter

from cadCAD.utils import flatten

VarDictType = Dict[str, List[Any]]
StatesListsType = List[Dict[str, Any]]
ConfigsType = List[Tuple[List[Callable], List[Callable]]]
EnvProcessesType = Dict[str, Callable]


def single_proc_exec(
    simulation_execs: List[Callable],
    var_dict_list: List[VarDictType],
    states_lists: List[StatesListsType],
    configs_structs: List[ConfigsType],
    env_processes_list: List[EnvProcessesType],
    Ts: List[range],
    SimIDs,
    Ns: List[int],
    ExpIDs: List[int],
    SubsetIDs,
    SubsetWindows,
    configured_n
):
    print(f'Execution Mode: single_threaded')
    params = [
        simulation_execs, states_lists, configs_structs, env_processes_list,
        Ts, SimIDs, Ns, SubsetIDs, SubsetWindows
    ]
    simulation_exec, states_list, config, env_processes, T, sim_id, N, subset_id, subset_window = list(
        map(lambda x: x.pop(), params)
    )
    result = simulation_exec(
        var_dict_list, states_list, config, env_processes, T, sim_id, N, subset_id, subset_window, configured_n
    )
    return flatten(result)


def parallelize_simulations(
    simulation_execs: List[Callable],
    var_dict_list: List[VarDictType],
    states_lists: List[StatesListsType],
    configs_structs: List[ConfigsType],
    env_processes_list: List[EnvProcessesType],
    Ts: List[range],
    SimIDs,
    Ns: List[int],
    ExpIDs: List[int],
    SubsetIDs,
    SubsetWindows,
    configured_n
):

    print(f'Execution Mode: parallelized')
    params = list(
        zip(
            simulation_execs, var_dict_list, states_lists, configs_structs, env_processes_list,
            Ts, SimIDs, Ns, SubsetIDs, SubsetWindows
        )
    )

    len_configs_structs = len(configs_structs)

    unique_runs = Counter(SimIDs)
    sim_count = max(unique_runs.values())
    highest_divisor = int(len_configs_structs / sim_count)

    new_configs_structs, new_params = [], []
    for count in range(sim_count):
        if count == 0:
            new_params.append(
                params[count: highest_divisor]
            )
            new_configs_structs.append(
                configs_structs[count: highest_divisor]
            )
        elif count > 0:
            new_params.append(
                params[count * highest_divisor: (count + 1) * highest_divisor]
            )
            new_configs_structs.append(
                configs_structs[count * highest_divisor: (count + 1) * highest_divisor]
            )


    def threaded_executor(params):
        tp = TPool()
        if len_configs_structs > 1:
            results = tp.map(
                lambda t: t[0](t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8], t[9], configured_n), params
            )
        else:
            t = params[0]
            results = t[0](t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8], t[9], configured_n)

        tp.close()
        return results

    pp = PPool()
    results = flatten(list(pp.map(lambda params: threaded_executor(params), new_params)))
    pp.close()
    pp.join()
    pp.clear()
    # pp.restart()

    return results


def local_simulations(
        simulation_execs: List[Callable],
        var_dict_list: List[VarDictType],
        states_lists: List[StatesListsType],
        configs_structs: List[ConfigsType],
        env_processes_list: List[EnvProcessesType],
        Ts: List[range],
        SimIDs,
        Ns: List[int],
        ExpIDs: List[int],
        SubsetIDs,
        SubsetWindows,
        configured_n
    ):
    print(f'SimIDs   : {SimIDs}')
    print(f'SubsetIDs: {SubsetIDs}')
    print(f'Ns       : {Ns}')
    print(f'ExpIDs   : {ExpIDs}')
    config_amt = len(configs_structs)
    try:
        _params = None
        if config_amt == 1:
            _params = var_dict_list[0]
            return single_proc_exec(
                simulation_execs, _params, states_lists, configs_structs, env_processes_list,
                Ts, SimIDs, Ns, ExpIDs, SubsetIDs, SubsetWindows, configured_n
            )
        elif config_amt > 1: # and config_amt < remote_threshold:
            _params = var_dict_list
            return parallelize_simulations(
                simulation_execs, _params, states_lists, configs_structs, env_processes_list,
                Ts, SimIDs, Ns, ExpIDs, SubsetIDs, SubsetWindows, configured_n
            )
    except ValueError:
        raise ValueError("\'sim_configs\' N must > 0")



from functools import reduce
from typing import Dict, List
from collections import defaultdict
from itertools import product
import warnings

import functools
import operator

from pandas import DataFrame


class SilentDF(DataFrame):
    def __repr__(self):
        return str(hex(id(DataFrame)))
        #"pandas.core.frame.DataFrame"


def append_dict(dict, new_dict):
    dict.update(new_dict)
    return dict


def arrange_cols(df, reverse=False):
    session_metrics = ['session_id', 'user_id', 'simulation_id', 'run_id']
    sys_metrics = ['run', 'timestep', 'substep']
    result_cols = list(set(df.columns) - set(session_metrics) - set(sys_metrics))
    result_cols.sort(reverse=reverse)
    return df[session_metrics + sys_metrics + result_cols]


class IndexCounter:
    def __init__(self):
        self.i = 0

    def __call__(self):
        self.i += 1
        return self.i


def compose(*functions):
    return reduce(lambda f, g: lambda x: f(g(x)), functions, lambda x: x)


def pipe(x):
    return x


def print_pipe(x):
    print(x)
    return x


def flattenDict(l):
    def tupalize(k, vs):
        l = []
        if isinstance(vs, list):
            for v in vs:
                l.append((k, v))
        else:
            l.append((k, vs))
        return l

    flat_list = [tupalize(k, vs) for k, vs in l.items()]
    flat_dict = [dict(items) for items in product(*flat_list)]
    return flat_dict


def flatten(l):
    if isinstance(l, list):
        return functools.reduce(operator.iconcat, l, [])
    elif isinstance(l, dict):
        return flattenDict(l)


def flatMap(f, collection):
    return flatten(list(map(f, collection)))


def dict_filter(dictionary, condition):
    return dict([(k, v) for k, v in dictionary.items() if condition(v)])


def get_max_dict_val_len(g: Dict[str, List[int]]) -> int:
    return len(max(g.values(), key=len))


def tabulate_dict(d: Dict[str, List[int]]) -> Dict[str, List[int]]:
    max_len = get_max_dict_val_len(d)
    _d = {}
    for k, vl in d.items():
        if len(vl) != max_len:
            _d[k] = vl + list([vl[-1]] * (max_len-1))
        else:
            _d[k] = vl

    return _d


def flatten_tabulated_dict(d: Dict[str, List[int]]) -> List[Dict[str, int]]:
    max_len = get_max_dict_val_len(d)
    dl = [{} for i in range(max_len)]

    for k, vl in d.items():
        for v, i in zip(vl, list(range(len(vl)))):
            dl[i][k] = v

    return dl


def contains_type(_collection, type):
    return any(isinstance(x, type) for x in _collection)


def drop_right(l, n):
    return l[:len(l) - n]


def key_filter(l, keyname):
    if (type(l) == list):
        return [v[keyname] for v in l]
        # Keeping support to dictionaries for backwards compatibility
        # Should be removed in the future
    warnings.warn(
        "The use of a dictionary to describe Partial State Update Blocks will be deprecated. Use a list instead.",
        FutureWarning)
    return [v[keyname] for k, v in l.items()]


def groupByKey(l):
    d = defaultdict(list)
    for key, value in l:
        d[key].append(value)
    return list(dict(d).items()).pop()


# @curried
def rename(new_name, f):
    f.__name__ = new_name
    return f


def curry_pot(f, *argv):
    sweep_ind = f.__name__[0:5] == 'sweep'
    arg_len = len(argv)
    if sweep_ind is True and arg_len == 4:
        return f(argv[0])(argv[1])(argv[2])(argv[3])
    elif sweep_ind is False and arg_len == 4:
        return f(argv[0], argv[1], argv[2], argv[3])
    elif sweep_ind is True and arg_len == 3:
        return f(argv[0])(argv[1])(argv[2])
    elif sweep_ind is False and arg_len == 3:
        return f(argv[0], argv[1], argv[2])
    else:
        raise TypeError('curry_pot() needs 3 or 4 positional arguments')



import pandas as pd
from tabulate import tabulate

from cadCAD.configuration.utils import config_sim
from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from cadCAD.configuration import Experiment
from cadCAD import configs

# Policies per Mechanism
def p1m1(_g, step, sH, s):
    return {'policy1': 1}
def p2m1(_g, step, sH, s):
    return {'policy2': 2}

def p1m2(_g, step, sH, s):
    return {'policy1': 2, 'policy2': 2}
def p2m2(_g, step, sH, s):
    return {'policy1': 2, 'policy2': 2}

def p1m3(_g, step, sH, s):
    return {'policy1': 1, 'policy2': 2, 'policy3': 3}
def p2m3(_g, step, sH, s):
    return {'policy1': 1, 'policy2': 2, 'policy3': 3}


# Internal States per Mechanism
def add(y, x):
    return lambda _g, step, sH, s, _input: (y, s[y] + x)

def policies(_g, step, sH, s, _input):
    y = 'policies'
    x = _input
    return (y, x)


# Genesis States
genesis_states = {
    'policies': {},
    's1': 0
}

variables = {
    's1': add('s1', 1),
    "policies": policies
}

psubs = {
    "m1": {
        "policies": {
            "p1": p1m1,
            "p2": p2m1
        },
        "variables": variables
    },
    "m2": {
        "policies": {
            "p1": p1m2,
            "p2": p2m2
        },
        "variables": variables
    },
    "m3": {
        "policies": {
            "p1": p1m3,
            "p2": p2m3
        },
        "variables": variables
    }
}


sim_config = config_sim(
    {
        "N": 1,
        "T": range(3),
    }
)

exp = Experiment()
exp.append_configs(
    sim_configs=sim_config,
    initial_state=genesis_states,
    partial_state_update_blocks=psubs,
    policy_ops=[lambda a, b: a + b, lambda y: y * 2] # Default: lambda a, b: a + b
)

exec_mode = ExecutionMode()
local_proc_ctx = ExecutionContext(context=exec_mode.local_mode)
run = Executor(exec_context=local_proc_ctx, configs=configs)

raw_result, tensor_field, sessions = run.execute()
result = pd.DataFrame(raw_result)

print()
print("Tensor Field:")
print(tabulate(tensor_field, headers='keys', tablefmt='psql'))
print("Output:")
print(tabulate(result, headers='keys', tablefmt='psql'))
print()



from cadCAD.configuration.utils import config_sim
import pandas as pd
from cadCAD.utils import SilentDF
from simulations.regression_tests.experiments import ext_ds_exp

df = SilentDF(pd.read_csv('simulations/external_data/output.csv'))


def query(s, df):
    return df[
            (df['run'] == s['run']) & (df['substep'] == s['substep']) & (df['timestep'] == s['timestep'])
        ].drop(columns=['run', 'substep', "timestep"])

def p1(_g, substep, sL, s, **kwargs):
    result_dict = query(s, df).to_dict()
    del result_dict["ds3"]
    return {k: list(v.values()).pop() for k, v in result_dict.items()}

def p2(_g, substep, sL, s, **kwargs):
    result_dict = query(s, df).to_dict()
    del result_dict["ds1"], result_dict["ds2"]
    return {k: list(v.values()).pop() for k, v in result_dict.items()}

#integrate_ext_dataset
def integrate_ext_dataset(_g, step, sL, s, _input, **kwargs):
    result_dict = query(s, df).to_dict()
    return 'external_data', {k: list(v.values()).pop() for k, v in result_dict.items()}

def increment(y, incr_by):
    return lambda _g, step, sL, s, _input, **kwargs: (y, s[y] + incr_by)
increment = increment('increment', 1)

def view_policies(_g, step, sL, s, _input, **kwargs):
    return 'policies', _input


external_data = {'ds1': None, 'ds2': None, 'ds3': None}
state_dict = {
    'increment': 0,
    'external_data': external_data,
    'policies': external_data
}


policies = {"p1": p1, "p2": p2}
states = {'increment': increment, 'external_data': integrate_ext_dataset, 'policies': view_policies}
PSUB = {'policies': policies, 'states': states}

# needs M1&2 need behaviors
partial_state_update_blocks = {
    'PSUB1': PSUB,
    'PSUB2': PSUB,
    'PSUB3': PSUB
}

sim_config = config_sim({
    "N": 2,
    "T": range(4)
})

ext_ds_exp.append_configs(
    sim_configs=sim_config,
    initial_state=state_dict,
    partial_state_update_blocks=partial_state_update_blocks,
    policy_ops=[lambda a, b: {**a, **b}]
)



from cadCAD.configuration.utils import config_sim, access_block
from simulations.regression_tests.experiments import hist_exp

policies, variables = {}, {}
exclusion_list = ['nonexsistant', 'last_x', '2nd_to_last_x', '3rd_to_last_x', '4th_to_last_x']

# Policies per Mechanism

# WARNING: DO NOT delete elements from sH
# state_history, target_field, psu_block_offset, exculsion_list
def last_update(_g, substep, sH, s, **kwargs):
    return {"last_x": access_block(
            state_history=sH,
            target_field="last_x",
            psu_block_offset=-1,
            exculsion_list=exclusion_list
        )
    }
policies["last_x"] = last_update

def second2last_update(_g, substep, sH, s, **kwargs):
    return {"2nd_to_last_x": access_block(sH, "2nd_to_last_x", -2, exclusion_list)}
policies["2nd_to_last_x"] = second2last_update


# Internal States per Mechanism

# WARNING: DO NOT delete elements from sH
def add(y, x):
    return lambda _g, substep, sH, s, _input, **kwargs: (y, s[y] + x)
variables['x'] = add('x', 1)

# last_partial_state_update_block
def nonexsistant(_g, substep, sH, s, _input, **kwargs):
    return 'nonexsistant', access_block(sH, "nonexsistant", 0, exclusion_list)
variables['nonexsistant'] = nonexsistant

# last_partial_state_update_block
def last_x(_g, substep, sH, s, _input, **kwargs):
    return 'last_x', _input["last_x"]
variables['last_x'] = last_x

# 2nd to last partial state update block
def second_to_last_x(_g, substep, sH, s, _input, **kwargs):
    return '2nd_to_last_x', _input["2nd_to_last_x"]
variables['2nd_to_last_x'] = second_to_last_x

# 3rd to last partial state update block
def third_to_last_x(_g, substep, sH, s, _input, **kwargs):
    return '3rd_to_last_x', access_block(sH, "3rd_to_last_x", -3, exclusion_list)
variables['3rd_to_last_x'] = third_to_last_x

# 4th to last partial state update block
def fourth_to_last_x(_g, substep, sH, s, _input, **kwargs):
    return '4th_to_last_x', access_block(sH, "4th_to_last_x", -4, exclusion_list)
variables['4th_to_last_x'] = fourth_to_last_x


genesis_states = {
    'x': 0,
    'nonexsistant': [],
    'last_x': [],
    '2nd_to_last_x': [],
    '3rd_to_last_x': [],
    '4th_to_last_x': []
}

PSUB = {
    "policies": policies,
    "variables": variables
}

partial_state_update_block = {
    "PSUB1": PSUB,
    "PSUB2": PSUB,
    "PSUB3": PSUB
}

sim_config = config_sim(
    {
        "N": 2,
        "T": range(3),
    }
)


hist_exp.append_configs(
    sim_configs=sim_config,
    initial_state=genesis_states,
    partial_state_update_blocks=partial_state_update_block
)



import unittest
import pandas as pd

from cadCAD import configs
from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from testing.generic_test import make_generic_test

exec_mode = ExecutionMode()
exec_ctx = ExecutionContext(context=exec_mode.single_mode)
run = Executor(exec_context=exec_ctx, configs=configs)

raw_result, tensor_field, _ = run.execute()
result = pd.DataFrame(raw_result)
expected_results = {
    (0, 1, 0, 0): {'x': 0, 'nonexsistant': [], 'last_x': [], '2nd_to_last_x': [], '3rd_to_last_x': [], '4th_to_last_x': []},
    (0, 1, 1, 1): {'x': 1,
                'nonexsistant': [],
                'last_x': [{'simulation': 0, 'x': 0, 'run': 1, 'substep': 0, 'timestep': 0}],
                '2nd_to_last_x': [],
                '3rd_to_last_x': [],
                '4th_to_last_x': []},
    (0, 1, 1, 2): {'x': 2,
                'nonexsistant': [],
                'last_x': [{'simulation': 0, 'x': 0, 'run': 1, 'substep': 0, 'timestep': 0}],
                '2nd_to_last_x': [],
                '3rd_to_last_x': [],
                '4th_to_last_x': []},
    (0, 1, 1, 3): {'x': 3,
                'nonexsistant': [],
                'last_x': [{'simulation': 0, 'x': 0, 'run': 1, 'substep': 0, 'timestep': 0}],
                '2nd_to_last_x': [],
                '3rd_to_last_x': [],
                '4th_to_last_x': []},
    (0, 1, 2, 1): {'x': 4,
                'nonexsistant': [],
                'last_x': [
                    {'simulation': 0, 'x': 1, 'run': 1, 'substep': 1, 'timestep': 1},
                    {'simulation': 0, 'x': 2, 'run': 1, 'substep': 2, 'timestep': 1},
                    {'simulation': 0, 'x': 3, 'run': 1, 'substep': 3, 'timestep': 1}
                ],
                '2nd_to_last_x': [{'simulation': 0, 'x': 0, 'run': 1, 'substep': 0, 'timestep': 0}],
                '3rd_to_last_x': [],
                '4th_to_last_x': []},
    (0, 1, 2, 2): {'x': 5,
                'nonexsistant': [],
                'last_x': [
                    {'simulation': 0, 'x': 1, 'run': 1, 'substep': 1, 'timestep': 1},
                    {'simulation': 0, 'x': 2, 'run': 1, 'substep': 2, 'timestep': 1},
                    {'simulation': 0, 'x': 3, 'run': 1, 'substep': 3, 'timestep': 1}
                ],
                '2nd_to_last_x': [{'simulation': 0, 'x': 0, 'run': 1, 'substep': 0, 'timestep': 0}],
                '3rd_to_last_x': [],
                '4th_to_last_x': []},
    (0, 1, 2, 3): {'x': 6,
                'nonexsistant': [],
                'last_x': [
                    {'simulation': 0, 'x': 1, 'run': 1, 'substep': 1, 'timestep': 1},
                    {'simulation': 0, 'x': 2, 'run': 1, 'substep': 2, 'timestep': 1},
                    {'simulation': 0, 'x': 3, 'run': 1, 'substep': 3, 'timestep': 1}
                ],
                '2nd_to_last_x': [{'simulation': 0, 'x': 0, 'run': 1, 'substep': 0, 'timestep': 0}],
                '3rd_to_last_x': [],
                '4th_to_last_x': []},
    (0, 1, 3, 1): {'x': 7,
                'nonexsistant': [],
                'last_x': [
                    {'simulation': 0, 'x': 4, 'run': 1, 'substep': 1, 'timestep': 2},
                    {'simulation': 0, 'x': 5, 'run': 1, 'substep': 2, 'timestep': 2},
                    {'simulation': 0, 'x': 6, 'run': 1, 'substep': 3, 'timestep': 2}
                ],
                '2nd_to_last_x': [
                    {'simulation': 0, 'x': 1, 'run': 1, 'substep': 1, 'timestep': 1},
                    {'simulation': 0, 'x': 2, 'run': 1, 'substep': 2, 'timestep': 1},
                    {'simulation': 0, 'x': 3, 'run': 1, 'substep': 3, 'timestep': 1}
                ],
                '3rd_to_last_x': [{'simulation': 0, 'x': 0, 'run': 1, 'substep': 0, 'timestep': 0}],
                '4th_to_last_x': []},
    (0, 1, 3, 2): {'x': 8,
                'nonexsistant': [],
                'last_x': [
                    {'simulation': 0, 'x': 4, 'run': 1, 'substep': 1, 'timestep': 2},
                    {'simulation': 0, 'x': 5, 'run': 1, 'substep': 2, 'timestep': 2},
                    {'simulation': 0, 'x': 6, 'run': 1, 'substep': 3, 'timestep': 2}
                ],
                '2nd_to_last_x': [
                    {'simulation': 0, 'x': 1, 'run': 1, 'substep': 1, 'timestep': 1},
                    {'simulation': 0, 'x': 2, 'run': 1, 'substep': 2, 'timestep': 1},
                    {'simulation': 0, 'x': 3, 'run': 1, 'substep': 3, 'timestep': 1}
                ],
                '3rd_to_last_x': [{'simulation': 0, 'x': 0, 'run': 1, 'substep': 0, 'timestep': 0}],
                '4th_to_last_x': []},
    (0, 1, 3, 3): {'x': 9,
                'nonexsistant': [],
                'last_x': [
                    {'simulation': 0, 'x': 4, 'run': 1, 'substep': 1, 'timestep': 2},
                    {'simulation': 0, 'x': 5, 'run': 1, 'substep': 2, 'timestep': 2},
                    {'simulation': 0, 'x': 6, 'run': 1, 'substep': 3, 'timestep': 2}
                ],
                '2nd_to_last_x': [
                    {'simulation': 0, 'x': 1, 'run': 1, 'substep': 1, 'timestep': 1},
                    {'simulation': 0, 'x': 2, 'run': 1, 'substep': 2, 'timestep': 1},
                    {'simulation': 0, 'x': 3, 'run': 1, 'substep': 3, 'timestep': 1}
                ],
                '3rd_to_last_x': [{'simulation': 0, 'x': 0, 'run': 1, 'substep': 0, 'timestep': 0}],
                '4th_to_last_x': []}
}


def row(a, b):
    return a == b
params = [
    ["historical_state_access", result, expected_results,
           ['x', 'nonexsistant', 'last_x', '2nd_to_last_x', '3rd_to_last_x', '4th_to_last_x'], [row]]
    ]


class GenericTest(make_generic_test(params)):
    pass


if __name__ == '__main__':
    unittest.main()



from cadCAD.configuration.utils import config_sim


# Policies per Mechanism
from testing.experiments import exp_policy_agg


def p1m1(_g, step, sL, s, **kwargs):
    return {'policy1': 1}
def p2m1(_g, step, sL, s, **kwargs):
    return {'policy2': 2}

def p1m2(_g, step, sL, s, **kwargs):
    return {'policy1': 2, 'policy2': 2}
def p2m2(_g, step, sL, s, **kwargs):
    return {'policy1': 2, 'policy2': 2}

def p1m3(_g, step, sL, s, **kwargs):
    return {'policy1': 1, 'policy2': 2, 'policy3': 3}
def p2m3(_g, step, sL, s, **kwargs):
    return {'policy1': 1, 'policy2': 2, 'policy3': 3}


# Internal States per Mechanism
def add(y, x):
    return lambda _g, step, sH, s, _input, **kwargs: (y, s[y] + x)

def policies(_g, step, sH, s, _input, **kwargs):
    y = 'policies'
    x = _input
    return (y, x)


# Genesis States
genesis_states = {
    'policies': {},
    's1': 0
}

variables = {
    's1': add('s1', 1),
    "policies": policies
}

partial_state_update_block = {
    "m1": {
        "policies": {
            "p1": p1m1,
            "p2": p2m1
        },
        "variables": variables
    },
    "m2": {
        "policies": {
            "p1": p1m2,
            "p2": p2m2
        },
        "variables": variables
    },
    "m3": {
        "policies": {
            "p1": p1m3,
            "p2": p2m3
        },
        "variables": variables
    }
}


sim_config = config_sim(
    {
        "N": 1,
        "T": range(3),
    }
)


exp_policy_agg.append_configs(
    sim_configs=sim_config,
    initial_state=genesis_states,
    partial_state_update_blocks=partial_state_update_block,
    policy_ops=[lambda a, b: a + b, lambda y: y * 2] # Default: lambda a, b: a + b
)



import unittest
from pprint import pprint

from parameterized import parameterized
from functools import reduce

from tabulate import tabulate


def generate_assertions_df(df, expected_results, target_cols, evaluations):
    test_names = []
    for eval_f in evaluations:
        def wrapped_eval(a, b):
            try:
                return eval_f(a, b)
            except KeyError:
                return True

        test_name = f"{eval_f.__name__}_test"
        test_names.append(test_name)
        df[test_name] = df.apply(
            lambda x: wrapped_eval(
                x.filter(items=target_cols).to_dict(),
                expected_results[(x['subset'], x['run'], x['timestep'], x['substep'])]
            ),
            axis=1
        )

    return df, test_names


def make_generic_test(params):
    class TestSequence(unittest.TestCase):
        # pprint(params)

        def generic_test(self, tested_df, expected_reults, test_name):
            erroneous = tested_df[(tested_df[test_name] == False)]
            print(tabulate(tested_df, headers='keys', tablefmt='psql'))

            if erroneous.empty is False:  # Or Entire df IS NOT erroneous
                for index, row in erroneous.iterrows():
                    # expected = expected_reults[(row['simulation'], row['run'], row['timestep'], row['substep'])]
                    expected = expected_reults[(row['subset'], row['run'], row['timestep'], row['substep'])]
                    unexpected = {f"invalid_{k}": expected[k] for k in expected if k in row and expected[k] != row[k]}

                    for key in unexpected.keys():
                        erroneous[key] = None
                        erroneous.at[index, key] = unexpected[key]
                # etc.

            self.assertTrue(reduce(lambda a, b: a and b, tested_df[test_name]))

        @parameterized.expand(params)
        def test_validation(self, name, result_df, expected_results, target_cols, evaluations):
            tested_df, test_names = generate_assertions_df(result_df, expected_results, target_cols, evaluations)

            for test_name in test_names:
                self.generic_test(tested_df, expected_results, test_name)

    return TestSequence



from pprint import pprint
from time import time
from typing import Callable, Dict, List, Any, Tuple

from cadCAD.utils import flatten
from cadCAD.utils.execution import print_exec_info
from cadCAD.configuration import Configuration, Processor
from cadCAD.configuration.utils import TensorFieldReport, configs_as_objs
from cadCAD.engine.simulation import Executor as SimExecutor
from cadCAD.engine.execution import single_proc_exec, parallelize_simulations, local_simulations
from cadCAD.configuration.utils import configs_as_dicts

VarDictType = Dict[str, List[Any]]
StatesListsType = List[Dict[str, Any]]
ConfigsType = List[Tuple[List[Callable], List[Callable]]]
EnvProcessesType = Dict[str, Callable]


class ExecutionMode:
    local_mode = 'local_proc'
    multi_mode = 'multi_proc'
    distributed = 'dist_proc'
    single_mode = 'single_proc'
    # Backwards compatible modes below
    single_proc = 'single_proc'
    multi_proc = 'multi_proc'


class ExecutionContext:
    def __init__(self, context=ExecutionMode.local_mode, method=None, additional_objs=None) -> None:
        self.name = context
        if context == 'local_proc':
            self.method = local_simulations
        elif context == 'single_proc':
            self.method = single_proc_exec
        elif context == 'multi_proc':
            self.method = parallelize_simulations
        elif context == 'dist_proc':
            def distroduce_proc(
                    simulation_execs, var_dict_list, states_lists, configs_structs, env_processes_list, Ts, RunIDs,
                    ExpIDs,
                    SubsetIDs,
                    SubsetWindows,
                    exec_method,
                    sc, additional_objs=additional_objs
            ):
                return method(
                    simulation_execs, var_dict_list, states_lists, configs_structs, env_processes_list, Ts, RunIDs,
                    ExpIDs,
                    SubsetIDs,
                    SubsetWindows,
                    exec_method,
                    sc, additional_objs
                )

            self.method = distroduce_proc


class Executor:
    def __init__(self,
             exec_context: ExecutionContext, configs: List[Configuration], spark_context=None
    ) -> None:
        self.sc = spark_context
        self.SimExecutor = SimExecutor
        self.exec_method = exec_context.method
        self.exec_context = exec_context.name
        self.configs = configs

    def execute(self) -> Tuple[Any, Any, Dict[str, Any]]:
        config_proc = Processor()
        create_tensor_field = TensorFieldReport(config_proc).create_tensor_field

        sessions = []
        var_dict_list, states_lists = [], []
        Ts, Ns, SimIDs, RunIDs = [], [], [], []
        ExpIDs, ExpWindows, SubsetIDs, SubsetWindows = [], [], [], []
        eps, configs_structs, env_processes_list = [], [], []
        partial_state_updates, sim_executors = [], []
        config_idx = 0

        print_exec_info(self.exec_context, configs_as_objs(self.configs))

        t1 = time()
        for x in self.configs:
            sessions.append(
                {
                    'user_id': x.user_id, 'experiment_id': x.experiment_id, 'session_id': x.session_id,
                    'simulation_id': x.simulation_id, 'run_id': x.run_id,
                    'subset_id': x.subset_id, 'subset_window': x.subset_window
                }
            )
            Ts.append(x.sim_config['T'])
            Ns.append(x.sim_config['N'])

            ExpIDs.append(x.experiment_id)
            ExpWindows.append(x.exp_window)
            SimIDs.append(x.simulation_id)
            RunIDs.append(x.run_id)
            SubsetIDs.append(x.subset_id)
            SubsetWindows.append(x.subset_window)

            var_dict_list.append(x.sim_config['M'])
            states_lists.append([x.initial_state])
            eps.append(list(x.exogenous_states.values()))
            configs_structs.append(config_proc.generate_config(x.initial_state, x.partial_state_updates, eps[config_idx]))
            env_processes_list.append(x.env_processes)
            partial_state_updates.append(x.partial_state_updates)
            sim_executors.append(SimExecutor(x.policy_ops).simulation)

            config_idx += 1

            # print(self.exec_context)
            # pprint(x.sim_config['M'])
            # exit()

        def get_final_dist_results(simulations, psus, eps, sessions):
            tensor_fields = [create_tensor_field(psu, ep) for psu, ep in list(zip(psus, eps))]
            return simulations, tensor_fields, sessions

        def get_final_results(simulations, psus, eps, sessions, remote_threshold):
            flat_timesteps, tensor_fields = [], []
            for sim_result, psu, ep in list(zip(simulations, psus, eps)):
                flat_timesteps.append(flatten(sim_result))
                tensor_fields.append(create_tensor_field(psu, ep))

            flat_simulations = flatten(flat_timesteps)
            if config_amt == 1:
                return simulations, tensor_fields, sessions
            elif (config_amt > 1): # and (config_amt < remote_threshold):
                return flat_simulations, tensor_fields, sessions

        remote_threshold = 100
        config_amt = len(self.configs)

        def auto_mode_switcher(config_amt):
            try:
                if config_amt == 1:
                    return ExecutionMode.single_mode, single_proc_exec
                elif (config_amt > 1): # and (config_amt < remote_threshold):
                    return ExecutionMode.multi_mode, parallelize_simulations
            except AttributeError:
                if config_amt < 1:
                    raise ValueError('N must be > 1!')
                # elif config_amt > remote_threshold:
                #     print('Remote Threshold is N=100. Use ExecutionMode.dist_proc if N >= 100')

        final_result = None
        # original_context = self.exec_context
        original_N = len(configs_as_dicts(self.configs))
        if self.exec_context != ExecutionMode.distributed:
            # Consider Legacy Support
            if self.exec_context != ExecutionMode.local_mode:
                self.exec_context, self.exec_method = auto_mode_switcher(config_amt)

            print("Execution Method: " + self.exec_method.__name__)
            simulations_results = self.exec_method(
                sim_executors, var_dict_list, states_lists, configs_structs, env_processes_list, Ts, SimIDs, RunIDs,
                ExpIDs, SubsetIDs, SubsetWindows, original_N
            )
            final_result = get_final_results(simulations_results, partial_state_updates, eps, sessions, remote_threshold)
        elif self.exec_context == ExecutionMode.distributed:
            print("Execution Method: " + self.exec_method.__name__)
            simulations_results = self.exec_method(
                sim_executors, var_dict_list, states_lists, configs_structs, env_processes_list, Ts,
                SimIDs, RunIDs, ExpIDs, SubsetIDs, SubsetWindows, original_N, self.sc
            )
            final_result = get_final_dist_results(simulations_results, partial_state_updates, eps, sessions)

        t2 = time()
        print(f"Total execution time: {t2 - t1 :.2f}s")

        return final_result



import numpy as np
from datetime import timedelta
import pprint

from cadCAD.configuration.utils import env_trigger, var_substep_trigger, config_sim, time_step, psub_list

from typing import Dict, List

from simulations.regression_tests.experiments import sweep_exp

pp = pprint.PrettyPrinter(indent=4)

seeds = {
    'z': np.random.RandomState(1),
    'a': np.random.RandomState(2),
    'b': np.random.RandomState(3),
    'c': np.random.RandomState(3)
}

# Optional
g: Dict[str, List[int]] = {
    'alpha': [1],
    # 'beta': [2],
    # 'gamma': [3],
    'beta': [2, 5],
    'gamma': [3, 4],
    'omega': [7]
}

psu_steps = ['m1', 'm2', 'm3']
system_substeps = len(psu_steps)
var_timestep_trigger = var_substep_trigger([0, system_substeps])
env_timestep_trigger = env_trigger(system_substeps)
env_process = {}
psu_block = {k: {"policies": {}, "variables": {}} for k in psu_steps}

# ['s1', 's2', 's3', 's4']
# Policies per Mechanism
def p1m1(_g, step, sL, s, **kwargs):
    return {'param1': 1}
psu_block['m1']['policies']['p1'] = p1m1

def p2m1(_g, step, sL, s, **kwargs):
    return {'param2': 4}
psu_block['m1']['policies']['p2'] = p2m1

def p1m2(_g, step, sL, s, **kwargs):
    return {'param1': 'a', 'param2': _g['beta']}
psu_block['m2']['policies']['p1'] = p1m2

def p2m2(_g, step, sL, s, **kwargs):
    return {'param1': 'b', 'param2': 0}
psu_block['m2']['policies']['p2'] = p2m2

def p1m3(_g, step, sL, s, **kwargs):
    return {'param1': np.array([10, 100])}
psu_block['m3']['policies']['p1'] = p1m3

def p2m3(_g, step, sL, s, **kwargs):
    return {'param1': np.array([20, 200])}
psu_block['m3']['policies']['p2'] = p2m3


# Internal States per Mechanism
def s1m1(_g, step, sL, s, _input, **kwargs):
    return 's1', 0
psu_block['m1']["variables"]['s1'] = s1m1

def s2m1(_g, step, sL, s, _input, **kwargs):
    return 's2', _g['beta']
psu_block['m1']["variables"]['s2'] = s2m1

def s1m2(_g, step, sL, s, _input, **kwargs):
    return 's1', _input['param2']
psu_block['m2']["variables"]['s1'] = s1m2

def s2m2(_g, step, sL, s, _input, **kwargs):
    return 's2', _input['param2']
psu_block['m2']["variables"]['s2'] = s2m2

def s1m3(_g, step, sL, s, _input, **kwargs):
    return 's1', 0
psu_block['m3']["variables"]['s1'] = s1m3

def s2m3(_g, step, sL, s, _input, **kwargs):
    return 's2', 0
psu_block['m3']["variables"]['s2'] = s2m3


# Exogenous States
def update_timestamp(_g, step, sL, s, _input, **kwargs):
    y = 'timestamp'
    return y, time_step(dt_str=s[y], dt_format='%Y-%m-%d %H:%M:%S', _timedelta=timedelta(days=0, minutes=0, seconds=1))
for m in ['m1','m2','m3']:
    # psu_block[m]["variables"]['timestamp'] = update_timestamp
    psu_block[m]["variables"]['timestamp'] = var_timestep_trigger(y='timestamp', f=update_timestamp)
    # psu_block[m]["variables"]['timestamp'] = var_trigger(
    #     y='timestamp', f=update_timestamp, pre_conditions={'substep': [0, system_substeps]}, cond_op=lambda a, b: a and b
    # )

proc_one_coef = 0.7
def es3(_g, step, sL, s, _input, **kwargs):
    return 's3', s['s3'] + proc_one_coef
# use `timestep_trigger` to update every ts
for m in ['m1','m2','m3']:
    psu_block[m]["variables"]['s3'] = var_timestep_trigger(y='s3', f=es3)


def es4(_g, step, sL, s, _input, **kwargs):
    return 's4', s['s4'] + _g['gamma']
for m in ['m1','m2','m3']:
    psu_block[m]["variables"]['s4'] = var_timestep_trigger(y='s4', f=es4)

# Genesis States
genesis_states = {
    's1': 0.0,
    's2': 0.0,
    's3': 1.0,
    's4': 1.0,
    'timestamp': '2018-10-01 15:16:24'
}


# Environment Process
env_process["s3"] = [lambda _g, x: _g['beta'], lambda _g, x: x + 1]
env_process["s4"] = env_timestep_trigger(trigger_field='timestep', trigger_vals=[5], funct_list=[lambda _g, x: _g['beta']])


# config_sim Necessary
sim_config = config_sim(
    {
        "N": 5,
        "T": range(2),
        "M": g, # Optional
    }
)

# New Convention
partial_state_update_blocks = psub_list(psu_block, psu_steps)
sweep_exp.append_configs(
    # user_id='user_a',
    sim_configs=sim_config,
    initial_state=genesis_states,
    seeds=seeds,
    env_processes=env_process,
    partial_state_update_blocks=partial_state_update_blocks
)


# print()
# print("Partial State Update Block:")
# pp.pprint(partial_state_update_blocks)
# print()
# print()



from pprint import pprint

import pandas as pd
from tabulate import tabulate

from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from documentation.examples import sys_model_A, sys_model_B
from cadCAD import configs

exec_mode = ExecutionMode()

# Single Process Execution using a Single System Model Configuration:
# sys_model_A
sys_model_A = [configs[0]]
local_proc_ctx = ExecutionContext(context=exec_mode.local_mode)
sys_model_A_simulation = Executor(exec_context=local_proc_ctx, configs=sys_model_A)

sys_model_A_raw_result, sys_model_A_tensor_field, sessions = sys_model_A_simulation.execute()
sys_model_A_result = pd.DataFrame(sys_model_A_raw_result)
print()
print("Tensor Field: sys_model_A")
print(tabulate(sys_model_A_tensor_field, headers='keys', tablefmt='psql'))
print("Result: System Events DataFrame")
print(tabulate(sys_model_A_result, headers='keys', tablefmt='psql'))
print()

# # Multiple Processes Execution using Multiple System Model Configurations:
# # sys_model_A & sys_model_B
multi_proc_ctx = ExecutionContext(context=exec_mode.multi_proc)
sys_model_AB_simulation = Executor(exec_context=multi_proc_ctx, configs=configs)

sys_model_AB_raw_result, sys_model_AB_tensor_field, sessions = sys_model_AB_simulation.execute()
print()
pprint(sys_model_AB_raw_result)
# sys_model_AB_result = pd.DataFrame(sys_model_AB_raw_result)
print()
print(f"Tensor Field:")
print(tabulate(sys_model_AB_tensor_field, headers='keys', tablefmt='psql'))
# print("Result: System Events DataFrame:")
# print(tabulate(sys_model_AB_result, headers='keys', tablefmt='psql'))
# print()



from setuptools import setup, find_packages

short_description = "cadCAD: a differential games based simulation software package for research, validation, and \
        Computer Aided Design of economic systems"

long_description = """
cadCAD (complex adaptive systems computer-aided design) is a python based, unified modeling framework for stochastic 
dynamical systems and differential games for research, validation, and Computer Aided Design of economic systems created 
by BlockScience. It is capable of modeling systems at all levels of abstraction from Agent Based Modeling (ABM) to 
System Dynamics (SD), and enabling smooth integration of computational social science simulations with empirical data 
science workflows.

An economic system is treated as a state-based model and defined through a set of endogenous and exogenous state 
variables which are updated through mechanisms and environmental processes, respectively. Behavioral models, which may 
be deterministic or stochastic, provide the evolution of the system within the action space of the mechanisms. 
Mathematical formulations of these economic games treat agent utility as derived from the state rather than direct from 
an action, creating a rich, dynamic modeling framework. Simulations may be run with a range of initial conditions and 
parameters for states, behaviors, mechanisms, and environmental processes to understand and visualize network behavior 
under various conditions. Support for A/B testing policies, Monte Carlo analysis, and other common numerical methods is 
provided.
"""

setup(name='cadCAD',
      version='0.4.21',
      description=short_description,
      long_description=long_description,
      url='https://github.com/cadCAD-org/cadCAD',
      author='Joshua E. Jodesty',
      author_email='joshua@block.science',
      license='LICENSE.txt',
      packages=find_packages(),
      install_requires=[
            "pandas",
            "pathos",
            "fn",
            "funcy"
      ],
      python_requires='>=3.6'
)



import pandas as pd
from tabulate import tabulate
from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from documentation.examples import sys_model_A, sys_model_B
from cadCAD import configs

exec_mode = ExecutionMode()

# Multiple Processes Execution using Multiple System Model Configurations:
local_proc_ctx = ExecutionContext(context=exec_mode.local_mode)
sys_model_AB_simulation = Executor(exec_context=local_proc_ctx, configs=configs)

i = 0
config_names = ['sys_model_A', 'sys_model_B']
sys_model_AB_raw_result, sys_model_AB_tensor_field, sessions = sys_model_AB_simulation.execute()
sys_model_AB_result = pd.DataFrame(sys_model_AB_raw_result)
print()
print(f"Tensor Field:")
print(tabulate(sys_model_AB_tensor_field, headers='keys', tablefmt='psql'))
print("Result: System Events DataFrame:")
print(tabulate(sys_model_AB_result, headers='keys', tablefmt='psql'))
print()



from pprint import pprint

import pandas as pd
from tabulate import tabulate
from cadCAD.engine import ExecutionMode, ExecutionContext, Executor

from testing.models import param_sweep
from cadCAD import configs

exec_mode = ExecutionMode()

exec_ctx = ExecutionContext(context=exec_mode.local_mode)
# exec_ctx = ExecutionContext(context=exec_mode.multi_proc)
run = Executor(exec_context=exec_ctx, configs=configs)

raw_result, tensor_fields, sessions = run.execute()
result = pd.DataFrame(raw_result)
# print(tabulate(tensor_fields[0], headers='keys', tablefmt='psql'))
# pprint(sessions)
print(tabulate(result, headers='keys', tablefmt='psql'))

print()

raw_result, tensor_fields, sessions = run.execute()
result = pd.DataFrame(raw_result)
# print(tabulate(tensor_fields[0], headers='keys', tablefmt='psql'))
# pprint(sessions)
print(tabulate(result, headers='keys', tablefmt='psql'))


from datetime import datetime


def datetime_range(start, end, delta, dt_format='%Y-%m-%d %H:%M:%S'):
    reverse_head = end
    [start, end] = [datetime.strptime(x, dt_format) for x in [start, end]]

    def _datetime_range(start, end, delta):
        current = start
        while current < end:
            yield current
            current += delta

    reverse_tail = [dt.strftime(dt_format) for dt in _datetime_range(start, end, delta)]
    return reverse_tail + [reverse_head]


def last_index(l):
    return len(l)-1


def retrieve_state(l, offset):
    return l[last_index(l) + offset + 1]


# @curried
def engine_exception(ErrorType, error_message, exception_function, try_function):
    try:
        return try_function
    except ErrorType:
        print(error_message)
        return exception_function


# @curried
# def fit_param(param, x):
#     return x + param



from pyspark.sql import DataFrame, Row
from tabulate import tabulate
from pprint import pprint
import pandas as pd

from simulations.regression_tests.models import config1, config2
from cadCAD.engine import ExecutionMode, ExecutionContext, Executor
from cadCAD.configuration.utils import configs_as_dataframe #, configs_as_objs, configs_as_dicts
from cadCAD.utils.sys_exec import to_spark_df, to_pandas_df
from cadCAD import configs

from distroduce.engine.execution import transform, distributed_simulations
from distroduce.session import sc_alt as sc

exec_mode = ExecutionMode()
distributed_sims = distributed_simulations(transform(publish=False))

distributed_ctx = ExecutionContext(context=exec_mode.distributed, method=distributed_sims)
run = Executor(exec_context=distributed_ctx, configs=configs, spark_context=sc)

raw_result, tensor_fields, sessions = run.execute()

print(tabulate(tensor_fields[0], headers='keys', tablefmt='psql'))
print()
# pprint(sessions)
# print()

print("Configuration Data:")
configs_df = configs_as_dataframe(configs)
print(tabulate(configs_df, headers='keys', tablefmt='psql'))
print("Tensor Field:")
print(tabulate(tensor_fields[0], headers='keys', tablefmt='psql'))
print("Output:")

# RDD:
print()
print("RDD:")
result: list = raw_result.take(5)
pprint(result[:2])
# to get all results execute the following
# result: list = raw_result.collect()
print()

print("Spark DataFrame:")
sdf: DataFrame = to_spark_df(raw_result)
# sdf: DataFrame = to_spark_df(raw_result, spark)
sdf.show(5)
print()

# Pandas:
print()
print("Pandas DataFrame:")
# pdf: pd.DataFrame = to_pandas_df(raw_result, config1.genesis_states)
pdf: pd.DataFrame = to_pandas_df(raw_result)
print(tabulate(pdf.head(), headers='keys', tablefmt='psql'))



from collections import namedtuple
from inspect import getmembers, ismethod
from pandas.core.frame import DataFrame

from cadCAD.utils import SilentDF


def val_switch(v):
    if isinstance(v, DataFrame) is True:
        return SilentDF(v)
    else:
        return v


class udcView(object):
    def __init__(self, d, masked_members):
        self.__dict__ = d
        self.masked_members = masked_members

    def __repr__(self):
        members = {}
        variables = {
            k: val_switch(v) for k, v in self.__dict__.items()
            if str(type(v)) != "<class 'method'>" and k not in self.masked_members # and isinstance(v, DataFrame) is not True
        }
        members['methods'] = [k for k, v in self.__dict__.items() if str(type(v)) == "<class 'method'>"]

        members.update(variables)
        return f"{members}"


class udcBroker(object):
    def __init__(self, obj, function_filter=['__init__']):
        d = {}
        funcs = dict(getmembers(obj, ismethod))
        filtered_functions = {k: v for k, v in funcs.items() if k not in function_filter}
        d['obj'] = obj
        d.update(vars(obj))  # somehow is enough
        d.update(filtered_functions)

        self.members_dict = d

    def get_members(self):
        return self.members_dict

    def get_view(self, masked_members):
        return udcView(self.members_dict, masked_members)

    def get_namedtuple(self):
        return namedtuple("Hydra", self.members_dict.keys())(*self.members_dict.values())


def UDO(udo, masked_members=['obj']):
    return udcBroker(udo).get_view(masked_members)


def udoPipe(obj_view):
    return UDO(obj_view.obj, obj_view.masked_members)



#
# def record_generator(row, cols):
#     return {col: row[col] for col in cols}

def gen_metric_row(row, cols):
    return ((row['run'], row['timestep'], row['substep']), {col: row[col] for col in cols})

# def gen_metric_row(row):
#     return ((row['run'], row['timestep'], row['substep']), {'s1': row['s1'], 'policies': row['policies']})

# def gen_metric_row(row):
#     return {
#         'run': row['run'],
#         'timestep': row['timestep'],
#         'substep': row['substep'],
#         's1': row['s1'],
#         'policies': row['policies']
#     }

def gen_metric_dict(df, cols):
    return dict([gen_metric_row(row, cols) for index, row in df.iterrows()])



from datetime import datetime
from kafka import KafkaProducer

def adjust_substep(substep):
    if substep == 0:
        new_substep = substep
    else:
        new_substep = substep - 1
    return new_substep

'''
Function that produces a single actions taken by a user and its accompanying message.

The results will be combined with a user-defined aggregation function (UDAF) 
given to the `policy_ops` argument of `cadCAD.configuration.append_configs`.
Current UDAF: `policy_ops=[lambda a, b: a + b]`
'''
def messages(client_id, room, action, _input, sender, receiver=None):
    return {
        'types': [action],
        'messages': [
            {
                'client': client_id, 'room': room, 'action': action,
                'sender': sender, 'receiver': receiver,
                'input': _input,
                'created': datetime.now()
            }
        ]
    }


'''
Policy representing a user entering a chat room and its accompanying message
'''
def enter_action(state, room, user):
    def f(_g, step, sL, s, **kwargs):
        msgs = messages(state, room, 'enter', f"{user} enters {room}", user)
        msgs['send_times'] = [0.000000]
        msgs['msg_counts'] = [len(msgs['messages'])]
        return msgs
    return f


'''
Policy representing a user sending a message to a receiver within a chat room
and its accompanying message

A Kafka Producer is used to send messages to a Kafka cluster. 
The configuration of the Kafka Producer via `cadCAD.engine.ExecutionContext`
'''
def message_actions(state, room, message_input, sender, receiver):
    msgs = messages(state, room, 'send', message_input, sender, receiver)
    msgs_list = msgs['messages']
    def send_action(_g, step, sL, s, kwargs_1):
        kafkaConfig = kwargs_1
        start_time = datetime.now()
        for msg in msgs_list:
            producer: KafkaProducer = kafkaConfig['producer']
            topic: str = kafkaConfig['send_topic']
            encoded_msg = str(msg).encode('utf-8')
            # part = adjust_substep(s['substep'])
            # print(f'PARTITION ------ {part}')
            # producer.send(topic=topic, value=encoded_msg, partition=part)
            producer.send(topic=topic, value=encoded_msg)
        msgs['send_times'] = [(datetime.now() - start_time).total_seconds()]
        msgs['msg_counts'] = [len(msgs_list)]
        return msgs

    return send_action


'''
Policy representing a user exiting a chat room and its accompanying message
'''
def exit_action(state, room, user):
    def f(_g, step, sL, s, **kwargs):
        msgs = messages(state, room, 'exit', f"{user} exited {room}", user)
        msgs_list = msgs['messages']
        msgs['send_times'] = [0.000000]
        msgs['msg_counts'] = [len(msgs_list)]
        return msgs
    return f



import warnings
from pprint import pprint

from pyspark import RDD, Row
from pyspark.sql import DataFrame, SparkSession
import pandas as pd

# Distributed
from tabulate import tabulate


def align_type(init_condition: dict):
    def f(d):
        for y, x in init_condition.items():
            d[y] = type(x)(d[y])
        return Row(**d)
    return f

### Typefull Conversion: to Spark
# rdd -> spark
### Typeless Conversion: to Spark
# (rdd -> pandas) -> spark
def to_spark(rdd: RDD, init_condition: dict):
    type_aligner = align_type(init_condition)
    rdd: RDD = rdd.map(type_aligner)
    return rdd.toDF()

### Typefull Conversion: to Pandas
# (rdd -> spark) -> pandas
### Typeless Conversion: to Pandas
# rdd -> pandas
def to_pandas(rdd: RDD):
    rdd_result: RDD = rdd.collect()
    pdf_from_rdd: DataFrame = pd.DataFrame(rdd_result)
    del rdd_result
    return pdf_from_rdd


def to_pandas_df(rdd: RDD, string_conversion=False, init_condition: dict = None):
    if init_condition is not None and string_conversion is False:
        # Typefull
        return to_spark(rdd=rdd, init_condition=init_condition).toPandas()
    elif init_condition is None and string_conversion is True:
        # String
        return rdd.map(lambda d: Row(**dict([(k, str(v)) for k, v in d.items()]))).toDF()
    else:
        # Typeless
        return to_pandas(rdd)


def to_spark_df(rdd: RDD, spark: SparkSession = None, init_condition: dict = None):
    if init_condition is not None and spark is not None:
        # Typefull
        return to_spark(rdd, init_condition)
    elif spark is None and init_condition is None:
        # String
        return rdd.map(lambda d: Row(**dict([(k, str(v)) for k, v in d.items()]))).toDF()
    else:
        # Typeless
        spark.conf.set("spark.sql.execution.arrow.enabled", "true")
        spark.conf.set("spark.sql.execution.arrow.fallback.enabled", "true")
        warnings.simplefilter(action='ignore', category=UserWarning)
        pdf_from_rdd: DataFrame = to_pandas(rdd)
        result = spark.createDataFrame(pdf_from_rdd)
        del pdf_from_rdd
        return result
